# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os
import unittest

import numpy
import torch
from bert_padding import pad_input, unpad_input
from einops import rearrange, repeat
from onnx import TensorProto, helper
from parameterized import parameterized
from test_gqa import attention_ref, has_flash_attention

from onnxruntime import InferenceSession, SessionOptions

torch.manual_seed(0)

pipeline_mode = True  # Reduces number of tests so pipeline doesn't time out


class Formats:
    BSNH = 0
    BNSH = 1


class Config:
    batch_size = 0
    sequence_length = 0
    kv_sequence_length = 0  # this is past sequence length when there is past state.
    num_heads = 0
    kv_num_heads = 0
    head_size = 0
    ep = "CUDAExecutionProvider"

    def __init__(self, batch_size, sequence_length, kv_sequence_length, num_heads, kv_num_heads, head_size):
        self.batch_size = batch_size
        self.sequence_length = sequence_length
        self.kv_sequence_length = kv_sequence_length
        self.num_heads = num_heads
        self.kv_num_heads = kv_num_heads
        self.head_size = head_size

    def __repr__(self):
        short_ep = self.ep[: -len("ExecutionProvider")].lower()
        return (
            f"Config(batch_size={self.batch_size}, sequence_length={self.sequence_length}, "
            f"kv_sequence_length={self.kv_sequence_length}, "
            f"num_heads={self.num_heads}, kv_num_heads={self.kv_num_heads}, head_size={self.head_size}, ep={short_ep})"
        )


def create_packed_multihead_attention_graph(config: Config):
    nodes = [
        helper.make_node(
            "PackedMultiHeadAttention",
            [
                "query",
                "",
                "",
                "",
                "token_offset",
                "cumulative_sequence_length",
            ],
            ["output"],
            "PackedMultiHeadAttention_0",
            num_heads=config.num_heads,
            domain="com.microsoft",
        ),
    ]

    graph = helper.make_graph(
        nodes,
        "PackedMultiHeadAttention_Graph",
        [
            helper.make_tensor_value_info(
                "query",
                TensorProto.FLOAT16,
                [
                    -1,
                    config.num_heads,
                    3,
                    config.head_size,
                ],
            ),
            helper.make_tensor_value_info(
                "token_offset", TensorProto.INT32, [config.batch_size, config.sequence_length]
            ),
            helper.make_tensor_value_info("cumulative_sequence_length", TensorProto.INT32, [config.batch_size + 1]),
        ],
        [
            helper.make_tensor_value_info(
                "output",
                TensorProto.FLOAT16,
                [-1, config.num_heads * config.head_size],
            ),
        ],
    )

    model = helper.make_model(graph)
    return model.SerializeToString()


def create_multihead_attention_graph(config: Config):
    nodes = [
        helper.make_node(
            "MultiHeadAttention",
            [
                "query",
                "key",
                "value",
            ],
            ["output"],
            "MultiHeadAttention_0",
            num_heads=config.num_heads,
            domain="com.microsoft",
        ),
    ]

    graph = helper.make_graph(
        nodes,
        "MultiHeadAttention_Graph",
        [
            helper.make_tensor_value_info(
                "query",
                TensorProto.FLOAT16,
                [
                    config.batch_size,
                    config.sequence_length,
                    config.num_heads * config.head_size,
                ],
            ),
            helper.make_tensor_value_info(
                "key",
                TensorProto.FLOAT16,
                [
                    config.batch_size,
                    config.kv_sequence_length,
                    config.num_heads * config.head_size,
                ],
            ),
            helper.make_tensor_value_info(
                "value",
                TensorProto.FLOAT16,
                [
                    config.batch_size,
                    config.kv_sequence_length,
                    config.num_heads * config.head_size,
                ],
            ),
        ],
        [
            helper.make_tensor_value_info(
                "output",
                TensorProto.FLOAT16,
                [config.batch_size, config.sequence_length, config.num_heads * config.head_size],
            ),
        ],
    )

    model = helper.make_model(graph)
    return model.SerializeToString()


def generate_random_padding_mask(max_seqlen, batch_size, device, mode="random"):
    assert mode in ["full", "random", "third"]
    if mode == "full":
        lengths = torch.full((batch_size, 1), max_seqlen, device=device, dtype=torch.int32)
    elif mode == "random":
        lengths = torch.randint(max(1, max_seqlen - 20), max_seqlen, (batch_size, 1), device=device)
    else:
        lengths = torch.randint(max_seqlen // 3, max_seqlen, (batch_size, 1), device=device)
    padding_mask = repeat(torch.arange(max_seqlen, device=device), "s -> b s", b=batch_size) < lengths
    return padding_mask


def generate_packed_qkv(q, k, v, query_padding_mask=None, key_padding_mask=None):
    """
    Arguments:
        q: (batch_size, seqlen_q, nheads, d)
        k: (batch_size, seqlen_k, nheads_k, d)
        v: (batch_size, seqlen_k, nheads_k, d)
        query_padding_mask: (batch_size, seqlen), bool
        key_padding_mask: (batch_size, seqlen), bool
    """
    batch_size, seqlen_q, nheads, d = q.shape
    _, seqlen_k, nheads_k, _ = k.shape
    assert k.shape == (batch_size, seqlen_k, nheads_k, d)
    assert v.shape == (batch_size, seqlen_k, nheads_k, d)

    if query_padding_mask is not None:
        q_unpad, indices_q, cu_seqlens_q, max_seqlen_q = unpad_input(q, query_padding_mask)

        def output_pad_fn(output_unpad):
            return pad_input(output_unpad, indices_q, batch_size, seqlen_q)

    else:
        q_unpad = rearrange(q, "b s h d -> (b s) h d")
        cu_seqlens_q = torch.arange(
            0, (batch_size + 1) * seqlen_q, step=seqlen_q, dtype=torch.int32, device=q_unpad.device
        )
        max_seqlen_q = seqlen_q

        def output_pad_fn(output_unpad):
            return rearrange(output_unpad, "(b s) h d -> b s h d", b=batch_size)

    if key_padding_mask is not None:
        k_unpad, _, _, _ = unpad_input(k, key_padding_mask)
        v_unpad, _, _, _ = unpad_input(v, key_padding_mask)
    else:
        k_unpad = rearrange(k, "b s h d -> (b s) h d")
        v_unpad = rearrange(v, "b s h d -> (b s) h d")

    assert (query_padding_mask == key_padding_mask).all()
    assert nheads == nheads_k
    qkv_unpad = torch.stack([q_unpad, k_unpad, v_unpad], dim=1)
    qkv = torch.stack([q, k, v], dim=2)
    if query_padding_mask is not None:

        def dqkv_pad_fn(dqkv_unpad):
            return pad_input(dqkv_unpad, indices_q, batch_size, seqlen_q)

    else:

        def dqkv_pad_fn(dqkv_unpad):
            return rearrange(dqkv_unpad, "(b s) t h d -> b s t h d", b=batch_size)

    return (
        qkv_unpad.detach().requires_grad_(),
        cu_seqlens_q,
        max_seqlen_q,
        qkv.detach().requires_grad_(),
        output_pad_fn,
        dqkv_pad_fn,
    )


def create_inputs(config: Config):
    qkv = torch.randn(
        config.batch_size,
        config.sequence_length,
        3,
        config.num_heads,
        config.head_size,
        device="cuda",
        dtype=torch.float16,
        requires_grad=False,
    )
    padding_mask = generate_random_padding_mask(config.sequence_length, config.batch_size, device="cuda", mode="random")
    qkv_unpad, cu_seqlens, max_seqlen, qkv, output_pad_fn, dqkv_pad_fn = generate_packed_qkv(
        *qkv.unbind(dim=2), padding_mask, padding_mask
    )
    return qkv_unpad, cu_seqlens, max_seqlen, qkv, output_pad_fn, dqkv_pad_fn, padding_mask


def generate_token_offset(cu_seqlens, max_seqlen):
    token_offset = []
    token_padset = []  # These are the indices that contain padding tokens
    for i in range(1, len(cu_seqlens)):
        start = i - 1
        pre_seqlen = cu_seqlens[i - 1]
        seqlen = cu_seqlens[i]
        token_offset += range(start * max_seqlen, (start * max_seqlen) + (seqlen - pre_seqlen))
        token_padset += range((start * max_seqlen) + (seqlen - pre_seqlen), i * max_seqlen)
    return numpy.asarray(token_offset + token_padset, dtype=numpy.int32)


def flash_attn_varlen_qkvpacked_func(qkv_unpad, cu_seqlens, token_offset, config):
    onnx_model_str = create_packed_multihead_attention_graph(config)
    qkv_unpad = torch.swapdims(qkv_unpad, 1, 2)
    ort_inputs = {
        "query": qkv_unpad.detach().cpu().numpy(),
        "token_offset": token_offset,
        "cumulative_sequence_length": cu_seqlens.cpu().numpy(),
    }
    sess_options = SessionOptions()
    ort_session = InferenceSession(onnx_model_str, sess_options, providers=[config.ep])
    ort_output = ort_session.run(None, ort_inputs)
    output = torch.tensor(ort_output)
    return output


def mha_func(q, k, v, config):
    onnx_model_str = create_multihead_attention_graph(config)
    q = torch.reshape(q, (config.batch_size, config.sequence_length, -1))
    k = torch.reshape(k, (config.batch_size, config.kv_sequence_length, -1))
    v = torch.reshape(v, (config.batch_size, config.kv_sequence_length, -1))
    ort_inputs = {
        "query": q.detach().cpu().numpy(),
        "key": k.detach().cpu().numpy(),
        "value": v.detach().cpu().numpy(),
    }
    sess_options = SessionOptions()
    ort_session = InferenceSession(onnx_model_str, sess_options, providers=[config.ep])
    ort_output = ort_session.run(None, ort_inputs)
    ort_output = numpy.array(ort_output)
    output = torch.tensor(ort_output)
    return output


def attention_qkvpacked_ref(
    qkv,
    key_padding_mask=None,
    causal=False,
    use_smooth_softmax=False,
):
    return attention_ref(
        qkv[:, :, 0],
        qkv[:, :, 1],
        qkv[:, :, 2],
        query_padding_mask=key_padding_mask,
        key_padding_mask=key_padding_mask,
        causal=causal,
        use_smooth_softmax=use_smooth_softmax,
    )


def parity_check_mha(
    config,
    packed,
    rtol=1e-3,
    atol=1e-3,
):
    if packed:
        qkv_unpad, cu_seqlens, _, qkv, output_pad_fn, _, key_padding_mask = create_inputs(config)
        token_offset = generate_token_offset(cu_seqlens, config.sequence_length).reshape(
            (config.batch_size, config.sequence_length)
        )
        # ORT Flash
        out_unpad = flash_attn_varlen_qkvpacked_func(qkv_unpad, cu_seqlens, token_offset, config)
        out_unpad = torch.squeeze(out_unpad, 0)
        out = torch.reshape(
            output_pad_fn(out_unpad), (config.batch_size, config.sequence_length, config.num_heads, config.head_size)
        )
        out = out.detach().cpu().numpy()
        # Pytorch to compare
        out_ref, _ = attention_qkvpacked_ref(qkv, key_padding_mask, causal=False)
        out_ref = out_ref.detach().cpu().numpy()
    else:
        q = torch.randn(
            config.batch_size,
            config.sequence_length,
            config.num_heads,
            config.head_size,
            device="cuda",
            dtype=torch.float16,
            requires_grad=False,
        )
        k = torch.randn(
            config.batch_size,
            config.kv_sequence_length,
            config.kv_num_heads,
            config.head_size,
            device="cuda",
            dtype=torch.float16,
            requires_grad=False,
        )
        v = torch.randn(
            config.batch_size,
            config.kv_sequence_length,
            config.kv_num_heads,
            config.head_size,
            device="cuda",
            dtype=torch.float16,
            requires_grad=False,
        )
        out = mha_func(q, k, v, config)
        out = torch.squeeze(out, 0)
        out = torch.reshape(out, (config.batch_size, config.sequence_length, config.num_heads, config.head_size))
        out = out.detach().cpu().numpy()
        # Pytorch to compare
        out_ref, _ = attention_ref(
            q, k, v, query_padding_mask=None, key_padding_mask=None, attention_bias=None, causal=False
        )
        out_ref = out_ref.detach().cpu().numpy()

    numpy.testing.assert_allclose(
        out, out_ref, rtol=rtol, atol=atol, equal_nan=True, err_msg=f" with {config} packed={packed}"
    )


def packed_mha_test_cases():
    batch_sizes = [2] if pipeline_mode else [1, 5]
    sequence_lengths = [1024, 1025] if pipeline_mode else [1024, 1025, 2048]
    num_heads = [1, 3] if pipeline_mode else [1, 6, 16]
    head_sizes = [16, 256] if pipeline_mode else [32, 40, 64, 80, 96, 128, 160, 192, 224, 256]

    for b in batch_sizes:
        for s in sequence_lengths:
            for n in num_heads:
                for h in head_sizes:
                    config = Config(b, s, s, n, n, h)
                    yield str(config), config


def mha_test_cases():
    batch_sizes = [2] if pipeline_mode else [1, 5]
    sequence_lengths = (
        [(1, 128), (113, 211), (2048, 2048)]
        if pipeline_mode
        else [
            (113, 203),
            (128, 217),
            (113, 211),
            (108, 256),
            (256, 512),
            (512, 256),
            (1024, 1024),
            (1023, 1024),
            (1024, 1023),
            (2048, 2048),
        ]
    )
    num_heads = [3] if pipeline_mode else [1, 6, 16]
    head_sizes = [64] if pipeline_mode else [32, 40, 64, 80, 96, 128, 160, 192, 224, 256]

    for b in batch_sizes:
        for s, kv_sequence_length in sequence_lengths:
            for n in num_heads:
                for h in head_sizes:
                    config = Config(b, s, kv_sequence_length, n, n, h)
                    yield str(config), config


class TestMHA(unittest.TestCase):
    @parameterized.expand(packed_mha_test_cases())
    def test_packed_mha(self, _, config):
        if not has_flash_attention():
            return
        os.environ["ORT_DISABLE_FLASH_ATTENTION"] = "0"
        print("-------- TEST PACKED MHA ---------")
        parity_check_mha(config, True)

    @parameterized.expand(mha_test_cases())
    def test_mha(self, _, config):
        if not has_flash_attention():
            return
        os.environ["ORT_DISABLE_FLASH_ATTENTION"] = "0"
        print("-------- TEST MHA ---------")
        parity_check_mha(config, False)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Generate SM120 FP4-based MoE GEMM kernel instantiations.

SM120 (Blackwell SM120) supports native FP4 grouped GEMM via CUTLASS TMA
warp-specialized kernels.  This script generates one .cu file per
(output_type, tile_shape) combination, each containing the
INSTANTIATE_TMA_WARP_SPECIALIZED_MOE_GEMM macro call.

Supported combinations:
  - FP4xFP4: activation=FP4, weight=FP4, output=fp16/bf16
    Tile shapes (element counts): 128x128x128, 128x128x256, 128x256x128, 256x128x128
    Uses NV-native nv_float4_t with ue4m3 SF (IsMXFPX=false).

  - FP8xFP4: activation=FP8 (e4m3), weight=FP4, output=fp16/bf16
    Tile shapes (element counts): 128x128x128
    Uses MX-format tuple with ue8m0 SF (IsMXFPX=true).
    Larger tiles (128x256, 256x128) exceed shared memory for >= 2 pipeline stages.
    K=64 is not supported (TMA tile size constraint for MXF8F6F4 on SM120).

Cluster shape: 1x1x1 only.
Fusion: NONE only (isValidSM120MOESpecialisation constraint).
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

GENERATED_PREFIX = "moe_gemm_tma_ws_sm120_fp4"
CLUSTER_M, CLUSTER_N, CLUSTER_K = 1, 1, 1


@dataclass(frozen=True, order=True)
class Instantiation:
    act_type_name: str  # "fp4" or "fp8"
    act_cpp_type: str  # "SafeFP4" or "SafeFP8"
    wt_cpp_type: str  # "SafeFP4"
    out_type_name: str  # "fp16" or "bf16"
    out_cpp_type: str  # "half" or "SafeBF16"
    m: int
    n: int
    k: int
    is_mxfpx: bool  # True for FP8xFP4

    @property
    def file_name(self) -> str:
        return (
            f"{GENERATED_PREFIX}_{self.act_type_name}_{self.out_type_name}_m{self.m}_n{self.n}_k{self.k}.generated.cu"
        )


def get_instantiations() -> list[Instantiation]:
    out_types = [
        ("fp16", "half"),
        ("bf16", "SafeBF16"),
    ]

    instantiations: set[Instantiation] = set()

    # FP4xFP4: all four tile shapes (K in FP4 elements)
    fp4_tile_shapes = [
        (128, 128, 128),
        (128, 128, 256),
        (128, 256, 128),
        (256, 128, 128),
    ]
    for out_name, out_cpp in out_types:
        for m, n, k in fp4_tile_shapes:
            instantiations.add(Instantiation("fp4", "SafeFP4", "SafeFP4", out_name, out_cpp, m, n, k, False))

    # FP8xFP4: only 128x128x128 fits in smem with >= 2 pipeline stages
    fp8_fp4_tile_shapes = [
        (128, 128, 128),
    ]
    for out_name, out_cpp in out_types:
        for m, n, k in fp8_fp4_tile_shapes:
            instantiations.add(Instantiation("fp8", "SafeFP8", "SafeFP4", out_name, out_cpp, m, n, k, True))

    return sorted(instantiations)


def render(inst: Instantiation) -> str:
    bf16_open = "#ifdef ENABLE_BF16\n" if inst.out_type_name == "bf16" else ""
    bf16_close = "#endif  // ENABLE_BF16\n" if inst.out_type_name == "bf16" else ""
    mxfpx_str = "true" if inst.is_mxfpx else "false"
    is_fp8 = inst.act_type_name == "fp8"
    fp8_open = "#ifdef ENABLE_FP8\n" if is_fp8 else ""
    fp8_close = "#endif  // ENABLE_FP8\n" if is_fp8 else ""

    return f"""/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 *
 * Generated by generate_moe_gemm_tma_ws_sm120_fp4.py. Do not edit manually.
 */

#ifndef EXCLUDE_SM_120
#ifdef COMPILE_BLACKWELL_SM120_TMA_GROUPED_GEMMS
#if defined(ENABLE_FP4) && defined(USE_FP4_QMOE)
{fp8_open}{bf16_open}
#include "contrib_ops/cuda/llm/moe_gemm/launchers/moe_gemm_tma_ws_launcher.inl"

namespace onnxruntime::llm::kernels::cutlass_kernels {{

INSTANTIATE_TMA_WARP_SPECIALIZED_MOE_GEMM(Sm120, {inst.act_cpp_type}, {inst.wt_cpp_type}, {inst.out_cpp_type}, EpilogueOpDefault, NONE, {inst.m}, {inst.n}, {inst.k}, {CLUSTER_M}, {CLUSTER_N}, {CLUSTER_K}, {mxfpx_str}, false)

}}  // namespace onnxruntime::llm::kernels::cutlass_kernels

{bf16_close}{fp8_close}#endif  // ENABLE_FP4 && USE_FP4_QMOE
#endif  // COMPILE_BLACKWELL_SM120_TMA_GROUPED_GEMMS
#endif  // EXCLUDE_SM_120
"""


def main() -> None:
    launcher_dir = Path(__file__).resolve().parent
    generated_names = {inst.file_name for inst in get_instantiations()}

    # Clean up stale generated files
    for generated_file in launcher_dir.glob(f"{GENERATED_PREFIX}_*.generated.cu"):
        if generated_file.name not in generated_names:
            generated_file.unlink()

    for inst in get_instantiations():
        path = launcher_dir / inst.file_name
        content = render(inst)
        if path.exists() and path.read_text(encoding="utf-8") == content:
            print(f"File {path.name} is up to date")
            continue
        path.write_text(content, encoding="utf-8")
        print(f"Generated {path.name}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
#
# Generate MoE GEMM kernels for SM80+:
#   python generate_moe_kernels.py -a "80;90" -o ./moe_gemm/launchers

import argparse
import os
from itertools import product

# CUDA type names for SM80 fused kernels.
CudaTypeName = {
    "bf16": "cutlass::bfloat16_t",
    "f16": "cutlass::half_t",
    "f32": "float",
}

# CUDA type names for SM90 TMA WS kernels (uses raw CUDA types)
CudaTypeNameSm90 = {
    "bf16": "SafeBF16",  # Alias defined in moe_gemm_tma_ws_launcher.inl
    "f16": "half",
    "f32": "float",
}

# Epilogue tags for SM80 fused kernels
EpilogueTagType = {
    "silu": "onnxruntime::llm::cutlass_extensions::EpilogueOpDefaultSilu",
    "gelu": "onnxruntime::llm::cutlass_extensions::EpilogueOpDefaultFtGelu",
}

# Epilogue tags for SM90 TMA WS kernels (must be EpilogueOpDefault for TMA WS)
EpilogueTagSm90 = {
    "default": "EpilogueOpDefault",
}

# Fusion types for SM90
FusionTypes = {
    "none": "NONE",
    "finalize": "FINALIZE",
}


def get_sm80_moe_template_instantiation(element_type, weight_type, tile_m, tile_n, tile_k, stages, epilogue_tag):
    """Generate a template instantiation for sm80_generic_fused_moe_gemm_kernelLauncher."""
    elem_cuda = CudaTypeName[element_type]
    weight_cuda = CudaTypeName[weight_type]
    epi_tag = EpilogueTagType[epilogue_tag]

    return f"""template void sm80_generic_fused_moe_gemm_kernelLauncher<{elem_cuda}, {weight_cuda}, {tile_m}, {tile_n}, {tile_k}, {stages}, {epi_tag}>(
    {elem_cuda} const*, {weight_cuda} const*, {elem_cuda} const*, bool, {elem_cuda}*,
    int64_t const*, int64_t, int64_t, int64_t, int, int, cudaStream_t, int*);
"""


def get_sm90_tma_ws_instantiation(
    arch_tag,
    dtype,
    weight_type,
    output_type,
    epi_tag,
    fusion,
    cta_m,
    cta_n,
    cta_k,
    cga_m,
    cga_n,
    cga_k,
    is_mxfpx,
    has_bias,
):
    """Generate an INSTANTIATE_TMA_WARP_SPECIALIZED_MOE_GEMM macro call for SM90+."""
    dtype_cuda = CudaTypeNameSm90[dtype]
    weight_cuda = CudaTypeNameSm90[weight_type]
    output_cuda = CudaTypeNameSm90[output_type]
    epi = EpilogueTagSm90[epi_tag]
    fuse = FusionTypes[fusion]
    mxfpx = "true" if is_mxfpx else "false"
    bias = "true" if has_bias else "false"

    return f"INSTANTIATE_TMA_WARP_SPECIALIZED_MOE_GEMM({arch_tag}, {dtype_cuda}, {weight_cuda}, {output_cuda}, {epi}, {fuse}, {cta_m}, {cta_n}, {cta_k}, {cga_m}, {cga_n}, {cga_k}, {mxfpx}, {bias})"


def generate_sm80_moe_operations():
    """Generate all SM80 MoE template instantiations."""
    operations = []

    # Data types: activation = weight (same type for fp16/bf16)
    data_types = [
        ("f16", "f16"),  # FP16
        ("bf16", "bf16"),  # BF16
    ]

    # Tile shapes: (MaxTileM, TileN, TileK)
    # From TensorRT-LLM generate_sm80_fused_grouped_gemm_operations():
    # cta_shapes_mnk = [(16, 128, 64), (16, 256, 64), (32, 128, 64), (64, 128, 64), (128, 128, 64)]
    # For gated activations TileN is halved internally, so 128->64, 256->128
    tile_shapes = [
        (16, 128, 64),
        (16, 256, 64),
        (32, 128, 64),
        (64, 128, 64),
        (128, 128, 64),
    ]

    # Stages from TRT-LLM: [2, 3, 4]
    stages_list = [2, 3, 4]

    # Epilogue tags for SwiGLU activation
    epilogue_tags = ["silu", "gelu"]

    for (elem_type, weight_type), (tile_m, tile_n, tile_k), stages, epi_tag in product(
        data_types, tile_shapes, stages_list, epilogue_tags
    ):
        operations.append(
            {
                "element_type": elem_type,
                "weight_type": weight_type,
                "tile_m": tile_m,
                "tile_n": tile_n,
                "tile_k": tile_k,
                "stages": stages,
                "epilogue_tag": epi_tag,
            }
        )

    return operations


def generate_sm90_tma_ws_operations():
    """Generate SM90 TMA Warp Specialized Grouped GEMM operations.

    Based on TensorRT-LLM's generate_sm90_grouped_gemm_operations().
    """
    operations = []

    # Data types
    data_types = [
        ("f16", "f16", "f16"),  # FP16
        ("bf16", "bf16", "bf16"),  # BF16
    ]

    # CTA shapes: M must be 128 for grouped GEMM
    # From TRT-LLM: M_TILES = [128], N_TILES = [16, 32, 64, 128, 256]
    m_tiles = [128]
    n_tiles = [16, 32, 64, 128, 256]
    cta_shapes_mn = [*list(product(m_tiles, n_tiles)), (256, 128)]

    # CGA (Cluster) shapes
    cga_shapes = list(product([1, 2], [1, 2], [1]))

    # Fusion types - SM90 supports fused finalize
    fusions = ["none", "finalize"]

    for (dtype, wtype, otype), (cta_m, cta_n), (cga_m, cga_n, cga_k), fusion in product(
        data_types, cta_shapes_mn, cga_shapes, fusions
    ):
        # Calculate K based on data type (128 bits / element size)
        bits_per_element = 16  # fp16 and bf16 are 16 bits
        cta_k = 128 * 8 // bits_per_element  # = 64

        operations.append(
            {
                "arch_tag": "Sm90",
                "dtype": dtype,
                "weight_type": wtype,
                "output_type": otype,
                "epi_tag": "default",
                "fusion": fusion,
                "cta_m": cta_m,
                "cta_n": cta_n,
                "cta_k": cta_k,
                "cga_m": cga_m,
                "cga_n": cga_n,
                "cga_k": cga_k,
                "is_mxfpx": False,
                "has_bias": False,
            }
        )

    return operations


def get_sm80_file_content(operations, arch):
    """Generate the content for a SM80 generated .cu file."""
    assert operations

    instantiations = []
    for op in operations:
        inst = get_sm80_moe_template_instantiation(
            op["element_type"],
            op["weight_type"],
            op["tile_m"],
            op["tile_n"],
            op["tile_k"],
            op["stages"],
            op["epilogue_tag"],
        )
        instantiations.append(inst)

    instantiation_block = "\n".join(instantiations)

    # Determine the exclusion guard based on arch
    exclude_macro = f"EXCLUDE_SM_{arch}"

    file_content = f"""/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 *
 * Auto-generated MoE GEMM kernel instantiations for SM{arch}.
 * DO NOT EDIT MANUALLY.
 */

#ifndef {exclude_macro}
#include "contrib_ops/cuda/llm/moe_gemm/launchers/fused_moe_gemm_launcher_sm80.inl"

namespace onnxruntime::llm::kernels::cutlass_kernels {{

#ifdef ENABLE_BF16
{instantiation_block}
#else
// BF16 not enabled, only instantiate FP16 variants
{get_fp16_only_instantiations(operations)}
#endif

}}  // namespace onnxruntime::llm::kernels::cutlass_kernels
#endif  // {exclude_macro}
"""
    return file_content


def get_sm90_file_content(operations, arch, dtype):
    """Generate the content for a SM90 TMA WS generated .cu file."""
    assert operations

    instantiations = []
    for op in operations:
        inst = get_sm90_tma_ws_instantiation(
            op["arch_tag"],
            op["dtype"],
            op["weight_type"],
            op["output_type"],
            op["epi_tag"],
            op["fusion"],
            op["cta_m"],
            op["cta_n"],
            op["cta_k"],
            op["cga_m"],
            op["cga_n"],
            op["cga_k"],
            op["is_mxfpx"],
            op["has_bias"],
        )
        instantiations.append(inst)

    instantiation_block = "\n".join(instantiations)

    file_content = f"""/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 *
 * Auto-generated SM90 TMA Warp Specialized Grouped GEMM instantiations for {dtype.upper()}.
 * DO NOT EDIT MANUALLY.
 */

#ifndef EXCLUDE_SM_90
#ifdef COMPILE_HOPPER_TMA_GROUPED_GEMMS

#include "contrib_ops/cuda/llm/moe_gemm/launchers/moe_gemm_tma_ws_launcher.inl"

namespace onnxruntime::llm::kernels::cutlass_kernels {{

{instantiation_block}

}}  // namespace onnxruntime::llm::kernels::cutlass_kernels

#endif  // COMPILE_HOPPER_TMA_GROUPED_GEMMS
#endif  // EXCLUDE_SM_90
"""
    return file_content


def get_fp16_only_instantiations(operations):
    """Generate instantiations for FP16 only."""
    fp16_ops = [op for op in operations if op["element_type"] == "f16"]
    instantiations = []
    for op in fp16_ops:
        inst = get_sm80_moe_template_instantiation(
            op["element_type"],
            op["weight_type"],
            op["tile_m"],
            op["tile_n"],
            op["tile_k"],
            op["stages"],
            op["epilogue_tag"],
        )
        instantiations.append(inst)
    return "\n".join(instantiations)


def write_file(content, output_file):
    """Write the generated content to a file."""
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Avoid changing modified time if file content is up to date
    if os.path.exists(output_file):
        with open(output_file) as f:
            if f.read() == content:
                print(f"File {output_file} is up to date")
                return

    with open(output_file, mode="w") as f:
        f.write(content)
    print(f"Generated {output_file}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate MoE GEMM kernel instantiations")
    parser.add_argument("-o", "--output_dir", type=str, required=True, help="Path to the output directory")
    parser.add_argument(
        "-a", "--architectures", type=str, required=True, help="Architectures to generate kernels for (e.g., '80;90')"
    )

    args = parser.parse_args()

    arches = args.architectures.split(";")
    output_dir = os.path.abspath(args.output_dir)

    def has_arch(sm):
        return f"{sm}" in arches or f"{sm}-real" in arches

    # Generate SM80 MoE kernels (fused gated activations)
    if has_arch(80) or has_arch(90):  # SM90 also uses SM80 kernels for non-TMA path
        operations = generate_sm80_moe_operations()

        # Group by element type for separate files to reduce compile time
        groups = {}
        for op in operations:
            key = op["element_type"]
            if key not in groups:
                groups[key] = []
            groups[key].append(op)

        for dtype, ops in groups.items():
            output_file = os.path.join(output_dir, f"fused_moe_gemm_sm80_{dtype}.generated.cu")
            content = get_sm80_file_content(ops, 80)
            write_file(content, output_file)

    # Generate SM90 TMA Warp Specialized Grouped GEMM kernels
    if has_arch(90):
        operations = generate_sm90_tma_ws_operations()

        # Group by dtype for separate files
        groups = {}
        for op in operations:
            key = op["dtype"]
            if key not in groups:
                groups[key] = []
            groups[key].append(op)

        for dtype, ops in groups.items():
            output_file = os.path.join(output_dir, f"moe_gemm_tma_ws_sm90_{dtype}.generated.cu")
            content = get_sm90_file_content(ops, 90, dtype)
            write_file(content, output_file)



import onnx
from onnx import TensorProto, helper


def create_model_with_node_output_not_used(model_path):
    # Create graph
    x = helper.make_tensor_value_info("X", TensorProto.FLOAT, [3, 2])
    w = helper.make_tensor_value_info("W", TensorProto.FLOAT, [2, 3])
    y = helper.make_tensor_value_info("Y", TensorProto.FLOAT, [2, 3])

    # Dropout node (two outputs)
    dropout_node = helper.make_node(
        "Dropout",
        inputs=["X"],
        outputs=["dropout_out", "dropout_mask"],
        name="DropoutNode",
    )

    # MatMul node
    matmul_node = helper.make_node(
        "MatMul",
        inputs=["dropout_out", "W"],
        outputs=["Y"],
        name="MatMulNode",
    )

    graph = helper.make_graph(
        nodes=[dropout_node, matmul_node],
        name="DropoutMatMulGraph",
        inputs=[x, w],
        outputs=[y],
    )

    model = helper.make_model(graph, opset_imports=[helper.make_operatorsetid("", 13)])

    onnx.checker.check_model(model)
    onnx.save(model, model_path)

    print(f"Model saved to: {model_path}")


if __name__ == "__main__":
    create_model_with_node_output_not_used("node_output_not_used.onnx")



#!/usr/bin/env python3
"""Build a wheel for the onnxruntime-ep-cuda12 or onnxruntime-ep-cuda13 package."""

import argparse
import platform
import re
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

SCRIPT_DIR = Path(__file__).parent
MIN_ONNXRUNTIME_VERSION_FILE = SCRIPT_DIR.parent / "MIN_ONNXRUNTIME_VERSION"

_TEMPLATE_VARIABLE_PATTERN = re.compile(r"@(\w+)@")
BINARY_PATTERNS = [
    "onnxruntime_providers_cuda_plugin.dll",
    "libonnxruntime_providers_cuda_plugin.so",
]
AUDITWHEEL_EXCLUDE = [
    "libcuda.so.1",
    "libcublas.so.12",
    "libcublas.so.13",
    "libcublasLt.so.12",
    "libcublasLt.so.13",
    "libcudart.so.12",
    "libcudart.so.13",
    "libcudnn.so.9",
    "libcufft.so.11",
    "libcufft.so.12",
    "libnvJitLink.so.12",
    "libnvJitLink.so.13",
    "libnvrtc.so.12",
    "libnvrtc.so.13",
    "libnvrtc-builtins.so.12",
    "libnvrtc-builtins.so.13",
]


def gen_file_from_template(template_file: Path, output_file: Path, variable_substitutions: dict[str, str]) -> None:
    content = template_file.read_text(encoding="utf-8")
    variables_in_file: set[str] = set()

    def replace(match: re.Match[str]) -> str:
        name = match.group(1)
        variables_in_file.add(name)
        return variable_substitutions.get(name, match.group(0))

    content = _TEMPLATE_VARIABLE_PATTERN.sub(replace, content)
    if variables_in_file != variable_substitutions.keys():
        provided = set(variable_substitutions.keys())
        raise ValueError(
            f"Template variables and substitution keys do not match for {template_file}. "
            f"Only in template: {sorted(variables_in_file - provided)}. "
            f"Only in substitutions: {sorted(provided - variables_in_file)}."
        )

    output_file.write_text(content, encoding="utf-8")


def prepare_staging_dir(staging_dir: Path, binary_dir: Path, version: str, package_name: str) -> None:
    staging_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy2(SCRIPT_DIR / "setup.py", staging_dir / "setup.py")
    shutil.copytree(SCRIPT_DIR / "onnxruntime_ep_cuda", staging_dir / "onnxruntime_ep_cuda")

    package_dir = staging_dir / "onnxruntime_ep_cuda"
    copied = []
    for pattern in BINARY_PATTERNS:
        for src in binary_dir.glob(pattern):
            dst = package_dir / src.name
            print(f"Copying {src} -> {dst}")
            shutil.copy2(src, dst)
            copied.append(dst)
    if not copied:
        raise FileNotFoundError(f"No plugin binaries found in {binary_dir}. Looked for: {BINARY_PATTERNS}")

    min_ort_version = MIN_ONNXRUNTIME_VERSION_FILE.read_text(encoding="utf-8").strip()
    if not min_ort_version:
        raise ValueError(f"{MIN_ONNXRUNTIME_VERSION_FILE} is empty")

    gen_file_from_template(
        SCRIPT_DIR / "pyproject.toml.in",
        staging_dir / "pyproject.toml",
        {"package_name": package_name, "version": version, "min_onnxruntime_version": min_ort_version},
    )


def build_wheel(source_dir: Path, wheel_dir: Path) -> None:
    wheel_dir.mkdir(parents=True, exist_ok=True)
    cmd = [
        sys.executable,
        "-m",
        "pip",
        "wheel",
        str(source_dir),
        "--wheel-dir",
        str(wheel_dir),
        "--no-deps",
        "--no-build-isolation",
    ]
    print(f"Running: {' '.join(cmd)}")
    subprocess.check_call(cmd)


def auditwheel_repair(wheel_dir: Path, wheel_name_prefix: str) -> None:
    if platform.system() != "Linux":
        return

    original_wheels = list(wheel_dir.glob(f"{wheel_name_prefix}-*.whl"))
    if not original_wheels:
        raise RuntimeError(f"No wheel found in {wheel_dir} to repair with auditwheel")

    with tempfile.TemporaryDirectory() as repaired_dir_name:
        repaired_dir = Path(repaired_dir_name)
        for wheel in original_wheels:
            cmd = [sys.executable, "-m", "auditwheel", "repair", str(wheel), "--wheel-dir", str(repaired_dir)]
            for lib in AUDITWHEEL_EXCLUDE:
                cmd.extend(["--exclude", lib])
            print(f"Running: {' '.join(cmd)}")
            subprocess.check_call(cmd)
            wheel.unlink()

        repaired_wheels = list(repaired_dir.glob("*.whl"))
        if not repaired_wheels:
            raise RuntimeError(f"auditwheel repair produced no wheels in {repaired_dir}")

        for repaired_wheel in repaired_wheels:
            repaired_wheel.replace(wheel_dir / repaired_wheel.name)


def collect_wheels(wheel_dir: Path, output_dir: Path, wheel_name_prefix: str) -> None:
    wheels = list(wheel_dir.glob(f"{wheel_name_prefix}-*.whl"))
    if not wheels:
        raise RuntimeError("No wheel was produced")
    output_dir.mkdir(parents=True, exist_ok=True)
    for wheel in wheels:
        dest = output_dir / wheel.name
        shutil.copy2(wheel, dest)
        print(f"Built wheel: {dest}")


def main() -> None:
    parser = argparse.ArgumentParser(description="Build onnxruntime-ep-cuda wheel")
    parser.add_argument("--binary_dir", required=True, type=Path, help="Directory containing built plugin EP binaries")
    parser.add_argument("--version", required=True, help="Package version string (PEP 440 format)")
    parser.add_argument("--package_name", required=True, help="Python distribution name to write into pyproject.toml")
    parser.add_argument("--output_dir", required=True, type=Path, help="Directory to place the built wheel")
    args = parser.parse_args()

    if not args.binary_dir.is_dir():
        raise FileNotFoundError(f"Binary directory does not exist: {args.binary_dir}")
    if not re.fullmatch(r"[A-Za-z0-9][A-Za-z0-9._-]*", args.package_name):
        raise ValueError(f"Invalid package name: {args.package_name}")

    wheel_name_prefix = args.package_name.replace("-", "_").replace(".", "_")

    with tempfile.TemporaryDirectory(prefix="ort_cuda_wheel_") as tmp:
        staging_dir = Path(tmp) / "package"
        wheel_dir = Path(tmp) / "wheels"
        prepare_staging_dir(staging_dir, args.binary_dir, args.version, args.package_name)
        build_wheel(staging_dir, wheel_dir)
        auditwheel_repair(wheel_dir, wheel_name_prefix)
        collect_wheels(wheel_dir, args.output_dir, wheel_name_prefix)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import onnx


def main():
    inp_shape = (1, 2, 8, 8)
    input_0 = onnx.helper.make_tensor_value_info("input_0", onnx.TensorProto.FLOAT, inp_shape)
    output_0 = onnx.helper.make_tensor_value_info("output_0", onnx.TensorProto.FLOAT, None)

    weight_data = [
        [[[-1.5, 0.0], [0.2, 1.5]], [[-1.5, 0.0], [0.2, 1.5]]],
        [[[-1.0, 0.0], [0.1333, 1.0]], [[-1.0, 0.0], [0.1333, 1.0]]],
    ]
    weight = onnx.numpy_helper.from_array(np.array(weight_data, dtype=np.float32), "weight")
    bias = onnx.numpy_helper.from_array(np.array([0.0, 0.0], dtype=np.float32), "bias")
    conv_node = onnx.helper.make_node("Conv", ["input_0", "weight", "bias"], ["output_0"], name="Conv0")
    graph = onnx.helper.make_graph(
        [conv_node],
        "Convf32",
        [input_0],
        [output_0],
        initializer=[weight, bias],
    )
    opset_imports = [onnx.helper.make_opsetid("", 21)]
    model = onnx.helper.make_model(graph, opset_imports=opset_imports)
    model = onnx.shape_inference.infer_shapes(model)

    onnx.checker.check_model(model, True)
    onnx.save_model(model, "conv_default_attrs.onnx")


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Shared utilities for the WebGPU plugin EP packaging scripts. Not a public API."""

from __future__ import annotations

import re
from pathlib import Path

# Matches "@var@" template variables.
_TEMPLATE_VARIABLE_PATTERN = re.compile(r"@(\w+)@")


def gen_file_from_template(
    template_file: Path, output_file: Path, variable_substitutions: dict[str, str], strict: bool = True
) -> None:
    """Generate a file from a template by substituting "@var@" markers with provided values.

    If `strict` is True, raises ValueError when the set of "@var@" names found in the template
    does not match the keys of `variable_substitutions`.

    Note: substituted values are inserted verbatim with no awareness of the target file's syntax.
    The caller is responsible for any quoting/escaping required by the target format.
    """
    content = template_file.read_text(encoding="utf-8")

    variables_in_file: set[str] = set()

    def replace(match: re.Match[str]) -> str:
        name = match.group(1)
        variables_in_file.add(name)
        return variable_substitutions.get(name, match.group(0))

    content = _TEMPLATE_VARIABLE_PATTERN.sub(replace, content)

    if strict and variables_in_file != variable_substitutions.keys():
        provided = set(variable_substitutions.keys())
        raise ValueError(
            f"Template variables and substitution keys do not match for {template_file}. "
            f"Only in template: {sorted(variables_in_file - provided)}. "
            f"Only in substitutions: {sorted(provided - variables_in_file)}."
        )

    output_file.write_text(content, encoding="utf-8")



from onnx import TensorProto, checker, helper, save, shape_inference

# A --> Squeeze --> Mul --> Relu --> Mul(2x) --> C
#                   ^
#                   |
# B ----------------+
graph_proto = helper.make_graph(
    nodes=[
        helper.make_node(
            "Squeeze",
            inputs=["A"],
            outputs=["squeeze0_output"],
            name="squeeze_0",
        ),
        helper.make_node(
            "Mul",
            inputs=["squeeze0_output", "B"],
            outputs=["mul0_output"],
            name="mul_0",
        ),
        helper.make_node(
            "Relu",
            inputs=["mul0_output"],
            outputs=["relu0_output"],
            name="relu_0",
        ),
        helper.make_node(
            "Mul",
            inputs=["relu0_output", "Const2"],
            outputs=["C"],
            name="mul_1",
        ),
    ],
    name="Main_graph",
    inputs=[
        helper.make_tensor_value_info("A", TensorProto.FLOAT, [3, 1, 2]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, [3, 2]),
    ],
    outputs=[
        helper.make_tensor_value_info("C", TensorProto.FLOAT, [3, 2]),
    ],
    initializer=[
        helper.make_tensor("Const2", TensorProto.FLOAT, [3, 2], [2.0] * 6),
    ],
)

model = helper.make_model(graph_proto)
model = shape_inference.infer_shapes(model)
checker.check_model(model, True)
save(model, "squeeze_mul_relu.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

"""
Profiling script for the CUDA QMoE GEMV decode path.

Usage:
  python profile_qmoe_gemv.py --case m1_top2_fp16_128x256 --warmup 5 --repeat 100

  nsys profile -t cuda,nvtx -o qmoe_gemv --export=sqlite \
      python profile_qmoe_gemv.py --case m1_top2_fp16_128x256 --warmup 5 --repeat 100
  python parse_nsys.py qmoe_gemv.sqlite --nvtx-range benchmark
"""

import argparse
import json
import os

import torch
from test_qmoe_cuda import (
    _QMOE_GEMV_BENCHMARK_RESULT_PREFIX,
    _qmoe_gemv_benchmark_case,
    _qmoe_gemv_benchmark_cases,
    run_qmoe_gemv_benchmark,
)


def _custom_case_from_args(args):
    case = dict(_qmoe_gemv_benchmark_case(args.case))
    custom_fields = {
        "batch_size": args.batch_size,
        "sequence_length": args.sequence_length,
        "hidden_size": args.hidden_size,
        "intermediate_size": args.intermediate_size,
        "num_experts": args.num_experts,
        "top_k": args.top_k,
        "onnx_dtype": args.dtype,
        "quant_bits": args.quant_bits,
        "block_size": args.block_size,
    }
    case.update({key: value for key, value in custom_fields.items() if value is not None})

    if any(value is not None for value in custom_fields.values()):
        case["name"] = (
            f"custom_m{case['batch_size'] * case['sequence_length']}_top{case['top_k']}_"
            f"{case['onnx_dtype'].lower()}_{case['hidden_size']}x{case['intermediate_size']}_"
            f"e{case['num_experts']}_int{case.get('quant_bits', 4)}_b{case.get('block_size', 0)}"
        )

    return case


def main():
    parser = argparse.ArgumentParser(description="Profile CUDA QMoE GEMV decode")
    parser.add_argument("--case", default="m1_top2_fp16_128x256", help="Benchmark case name")
    parser.add_argument("--list-cases", action="store_true", help="List available benchmark case names and exit")
    parser.add_argument("--batch-size", type=int, help="Override batch size")
    parser.add_argument("--sequence-length", type=int, help="Override sequence length")
    parser.add_argument("--hidden-size", type=int, help="Override hidden size")
    parser.add_argument("--intermediate-size", type=int, help="Override intermediate size")
    parser.add_argument("--num-experts", type=int, help="Override number of experts")
    parser.add_argument("--top-k", type=int, help="Override top-k experts per token")
    parser.add_argument("--dtype", choices=["FLOAT16", "BFLOAT16"], help="Override ONNX dtype")
    parser.add_argument("--quant-bits", type=int, choices=[4, 8], help="Override QMoE integer weight bits")
    parser.add_argument("--block-size", type=int, choices=[0, 32, 64, 128], help="Override QMoE INT block size")
    parser.add_argument("--warmup", type=int, default=5, help="Warmup iterations before the benchmark NVTX range")
    parser.add_argument("--repeat", type=int, default=100, help="Benchmark iterations")
    parser.add_argument(
        "--disable-gemv",
        action="store_true",
        help="Run the grouped GEMM fallback by setting ORT_DISABLE_MOE_GEMV=1 before session creation",
    )
    parser.add_argument(
        "--nvtx",
        action="store_true",
        help="Wrap the measured loop in an NVTX range named 'benchmark'",
    )
    args = parser.parse_args()

    if args.list_cases:
        for case in _qmoe_gemv_benchmark_cases():
            print(case["name"])
        return

    case = _custom_case_from_args(args)

    if not torch.cuda.is_available():
        raise RuntimeError("CUDA is required for QMoE GEMV profiling")

    os.environ["ORT_QMOE_GEMV_BENCHMARK_REPEATS"] = str(max(1, args.repeat))
    os.environ["ORT_QMOE_GEMV_BENCHMARK_WARMUP"] = str(max(0, args.warmup))
    if args.nvtx:
        os.environ["ORT_QMOE_GEMV_BENCHMARK_NVTX"] = "1"
    else:
        os.environ.pop("ORT_QMOE_GEMV_BENCHMARK_NVTX", None)

    if args.disable_gemv:
        os.environ["ORT_DISABLE_MOE_GEMV"] = "1"
    else:
        os.environ.pop("ORT_DISABLE_MOE_GEMV", None)

    result = run_qmoe_gemv_benchmark(case)
    if result["has_invalid_output"]:
        raise RuntimeError("QMoE GEMV profiling produced NaN or Inf output")

    print(_QMOE_GEMV_BENCHMARK_RESULT_PREFIX + json.dumps(result, sort_keys=True))


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Generator for synthetic BART SDPA attention ONNX graphs used in fusion tests.

This module reproduces the SDPA attention pattern emitted by HuggingFace
Transformers >= 4.49 when exporting BART models, so that the
FusionBartAttention pass can be exercised without a real model checkpoint.
"""

import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper


def create_bart_attention_sdpa(hidden_size: int = 16, num_heads: int = 4, with_mask: bool = True) -> onnx.ModelProto:
    """Create a minimal BART SDPA attention graph for fusion testing.

    The graph reproduces the self-attention subgraph exported by HuggingFace
    Transformers >= 4.49 for BART, including:
      - Pre-LayerNorm on the input
      - Q/K/V linear projections (MatMul + Add + Reshape + Transpose)
      - SDPA-specific K^T chain (Reshape -> Transpose(0,2,1) -> Reshape)
      - Separate Q and K scaling (Mul by 1/sqrt(head_dim))
      - QK MatMul, optional mask Add, Softmax, IsNaN/Where NaN guard
      - Attention * V MatMul
      - Output projection (MatMul + Add) with residual Add
      - Final LayerNormalization as the fusion anchor

    Args:
        hidden_size: Total hidden dimension (must be divisible by num_heads).
        num_heads: Number of attention heads.
        with_mask: If True, include an additive float attention mask input.

    Returns:
        An onnx.ModelProto representing the attention subgraph.
    """
    assert hidden_size % num_heads == 0, "hidden_size must be divisible by num_heads"

    head_dim = hidden_size // num_heads
    batch = 1
    seq = 8
    sqrt_scale = float(1.0 / (head_dim**0.5))

    # ------------------------------------------------------------------
    # Initializers (weights and shape constants)
    # ------------------------------------------------------------------
    np.random.seed(42)

    ln_weight = numpy_helper.from_array(np.ones(hidden_size, dtype=np.float32), "ln_weight")
    ln_bias = numpy_helper.from_array(np.zeros(hidden_size, dtype=np.float32), "ln_bias")

    q_weight = numpy_helper.from_array(np.random.randn(hidden_size, hidden_size).astype(np.float32), "q_weight")
    q_bias = numpy_helper.from_array(np.random.randn(hidden_size).astype(np.float32), "q_bias")
    k_weight = numpy_helper.from_array(np.random.randn(hidden_size, hidden_size).astype(np.float32), "k_weight")
    k_bias = numpy_helper.from_array(np.random.randn(hidden_size).astype(np.float32), "k_bias")
    v_weight = numpy_helper.from_array(np.random.randn(hidden_size, hidden_size).astype(np.float32), "v_weight")
    v_bias = numpy_helper.from_array(np.random.randn(hidden_size).astype(np.float32), "v_bias")

    out_weight = numpy_helper.from_array(np.random.randn(hidden_size, hidden_size).astype(np.float32), "out_weight")
    out_bias = numpy_helper.from_array(np.random.randn(hidden_size).astype(np.float32), "out_bias")

    ln2_weight = numpy_helper.from_array(np.ones(hidden_size, dtype=np.float32), "ln2_weight")
    ln2_bias = numpy_helper.from_array(np.zeros(hidden_size, dtype=np.float32), "ln2_bias")

    # Shape constants used by Reshape nodes.
    # Q/K/V projection reshape: [batch, seq, num_heads, head_dim]
    shape_qkv_4d = numpy_helper.from_array(np.array([batch, seq, num_heads, head_dim], dtype=np.int64), "shape_qkv_4d")
    # K^T chain — first Reshape merges batch and num_heads dims:
    #   [batch, num_heads, seq, head_dim] -> [batch*num_heads, seq, head_dim]
    shape_k_3d = numpy_helper.from_array(np.array([batch * num_heads, seq, head_dim], dtype=np.int64), "shape_k_3d")
    # K^T chain — second Reshape expands back to 4-D with transposed inner dims:
    #   [batch*num_heads, head_dim, seq] -> [batch, num_heads, head_dim, seq]
    shape_kt_4d = numpy_helper.from_array(np.array([batch, num_heads, head_dim, seq], dtype=np.int64), "shape_kt_4d")
    # Output reshape: [batch, seq, hidden_size]
    shape_output = numpy_helper.from_array(np.array([batch, seq, hidden_size], dtype=np.int64), "shape_output")

    # Scalar attention scale: 1 / sqrt(head_dim)
    scale_val = numpy_helper.from_array(np.array(sqrt_scale, dtype=np.float32), "sqrt_scale")

    # Constant used by the NaN guard Where node.
    zero_val = numpy_helper.from_array(np.array(0.0, dtype=np.float32), "zero_constant")

    # Large negative value for masked positions in attention mask.
    neg_inf_val = numpy_helper.from_array(np.array(-1e9, dtype=np.float32), "neg_inf_constant")

    initializers = [
        ln_weight,
        ln_bias,
        q_weight,
        q_bias,
        k_weight,
        k_bias,
        v_weight,
        v_bias,
        out_weight,
        out_bias,
        ln2_weight,
        ln2_bias,
        shape_qkv_4d,
        shape_k_3d,
        shape_kt_4d,
        shape_output,
        scale_val,
        zero_val,
        neg_inf_val,
    ]

    # ------------------------------------------------------------------
    # Graph inputs
    # ------------------------------------------------------------------
    input_1 = helper.make_tensor_value_info("input_1", TensorProto.FLOAT, [batch, seq, hidden_size])
    graph_inputs = [input_1]
    if with_mask:
        # Boolean mask input — the Where node converts it to a float mask,
        # matching the mask_nodes_bart pattern in fusion_bart_attention.py.
        attention_mask_bool = helper.make_tensor_value_info(
            "attention_mask_bool", TensorProto.BOOL, [batch, 1, seq, seq]
        )
        graph_inputs.append(attention_mask_bool)

    # ------------------------------------------------------------------
    # Graph output
    # ------------------------------------------------------------------
    graph_output = helper.make_tensor_value_info("output", TensorProto.FLOAT, [batch, seq, hidden_size])

    # ------------------------------------------------------------------
    # Nodes
    # ------------------------------------------------------------------
    nodes = []

    # 1. Pre-LayerNorm on the raw input.
    nodes.append(
        helper.make_node(
            "LayerNormalization",
            ["input_1", "ln_weight", "ln_bias"],
            ["layer_norm_out"],
            "layer_norm",
            axis=-1,
            epsilon=1e-5,
        )
    )

    # 2. Q projection: MatMul -> Add -> Reshape -> Transpose(0,2,1,3)
    nodes.append(helper.make_node("MatMul", ["layer_norm_out", "q_weight"], ["q_matmul_out"], "q_matmul"))
    nodes.append(helper.make_node("Add", ["q_matmul_out", "q_bias"], ["q_add_out"], "q_add"))
    nodes.append(helper.make_node("Reshape", ["q_add_out", "shape_qkv_4d"], ["q_reshape_out"], "q_reshape"))
    nodes.append(helper.make_node("Transpose", ["q_reshape_out"], ["q_transposed"], "q_transpose", perm=[0, 2, 1, 3]))

    # 3. K projection: MatMul -> Add -> Reshape -> Transpose(0,2,1,3)
    nodes.append(helper.make_node("MatMul", ["layer_norm_out", "k_weight"], ["k_matmul_out"], "k_matmul"))
    nodes.append(helper.make_node("Add", ["k_matmul_out", "k_bias"], ["k_add_out"], "k_add"))
    nodes.append(helper.make_node("Reshape", ["k_add_out", "shape_qkv_4d"], ["k_reshape_out"], "k_reshape"))
    nodes.append(helper.make_node("Transpose", ["k_reshape_out"], ["k_transposed"], "k_transpose", perm=[0, 2, 1, 3]))

    # 4. V projection: MatMul -> Add -> Reshape -> Transpose(0,2,1,3)
    nodes.append(helper.make_node("MatMul", ["layer_norm_out", "v_weight"], ["v_matmul_out"], "v_matmul"))
    nodes.append(helper.make_node("Add", ["v_matmul_out", "v_bias"], ["v_add_out"], "v_add"))
    nodes.append(helper.make_node("Reshape", ["v_add_out", "shape_qkv_4d"], ["v_reshape_out"], "v_reshape"))
    nodes.append(helper.make_node("Transpose", ["v_reshape_out"], ["v_transposed"], "v_transpose", perm=[0, 2, 1, 3]))

    # 5. SDPA-specific K^T chain.
    #
    #   k_transposed  [batch, num_heads, seq, head_dim]
    #     -> Reshape  [batch*num_heads, seq, head_dim]         (k_3d)
    #     -> Transpose(0,2,1)  [batch*num_heads, head_dim, seq] (k_3d_t)
    #     -> Reshape  [batch, num_heads, head_dim, seq]         (k_4d_t)
    #
    # The fusion pattern (k_nodes_sdpa) matches the data path:
    #   Mul <- Reshape <- Transpose <- Reshape <- Transpose <- Reshape <- Add <- MatMul
    # The two Reshapes here (k_3d, k_4d_t) are nodes 3 and 5 in that chain
    # (counting from the Mul), preceded by the initial K projection Reshape.
    nodes.append(helper.make_node("Reshape", ["k_transposed", "shape_k_3d"], ["k_3d"], "k_reshape_3d"))
    nodes.append(helper.make_node("Transpose", ["k_3d"], ["k_3d_t"], "k_transpose_3d", perm=[0, 2, 1]))
    nodes.append(helper.make_node("Reshape", ["k_3d_t", "shape_kt_4d"], ["k_4d_t"], "k_reshape_4d"))

    # 6. Separate Q and K scaling by 1/sqrt(head_dim).
    nodes.append(helper.make_node("Mul", ["q_transposed", "sqrt_scale"], ["q_scaled"], "q_scale"))
    nodes.append(helper.make_node("Mul", ["k_4d_t", "sqrt_scale"], ["k_scaled"], "k_scale"))

    # 7. QK attention scores.
    nodes.append(helper.make_node("MatMul", ["q_scaled", "k_scaled"], ["qk_matmul_out"], "qk_matmul"))

    # 8. Optional additive attention mask.
    #    In BART, the boolean mask is converted to a float mask via
    #    Where(condition, 0.0, -1e9) and then added to QK scores.
    #    The fusion code (mask_nodes_bart) matches: add_qk input[1] -> Where.
    if with_mask:
        nodes.append(
            helper.make_node(
                "Where",
                ["attention_mask_bool", "zero_constant", "neg_inf_constant"],
                ["attention_mask_float"],
                "mask_where",
            )
        )
        nodes.append(helper.make_node("Add", ["qk_matmul_out", "attention_mask_float"], ["qk_masked"], "qk_mask_add"))
        softmax_input = "qk_masked"
    else:
        softmax_input = "qk_matmul_out"

    # 9. Softmax over the last axis.
    nodes.append(helper.make_node("Softmax", [softmax_input], ["softmax_out"], "softmax", axis=-1))

    # 10. NaN guard: Where(IsNaN(softmax), 0, softmax).
    #     Where inputs: [condition, value_if_true, value_if_false]
    #     softmax_out is at index 2 (value_if_false), matching the fusion
    #     pattern qk_nodes_sdpa_{no,with}_mask which follows input[2].
    nodes.append(helper.make_node("IsNaN", ["softmax_out"], ["is_nan"], "isnan"))
    nodes.append(helper.make_node("Where", ["is_nan", "zero_constant", "softmax_out"], ["nan_guarded"], "nan_guard"))

    # 11. Attention output: NaN-guarded weights * V.
    nodes.append(helper.make_node("MatMul", ["nan_guarded", "v_transposed"], ["attn_v_out"], "attn_v_matmul"))

    # 12. Reshape attention output back to [batch, seq, hidden_size].
    nodes.append(
        helper.make_node("Transpose", ["attn_v_out"], ["attn_transposed"], "attn_transpose", perm=[0, 2, 1, 3])
    )
    nodes.append(helper.make_node("Reshape", ["attn_transposed", "shape_output"], ["attn_reshaped"], "attn_reshape"))

    # 13. Output projection.
    nodes.append(helper.make_node("MatMul", ["attn_reshaped", "out_weight"], ["out_matmul_out"], "out_matmul"))
    nodes.append(helper.make_node("Add", ["out_matmul_out", "out_bias"], ["out_add_out"], "out_add"))

    # 14. Residual connection.
    #
    #   We use layer_norm_out (a node output) rather than input_1 (a graph
    #   input) so that the fusion code can resolve root_input via
    #   output_name_to_node.  The first LayerNorm output also has Q/K/V
    #   MatMuls as direct children, which satisfies the fusion's heuristic
    #   for confirming the true attention root (lines 97-104 of
    #   fusion_bart_attention.py).
    nodes.append(helper.make_node("Add", ["layer_norm_out", "out_add_out"], ["residual_out"], "residual_add"))

    # 15. Final LayerNormalization — the fusion anchor node.
    nodes.append(
        helper.make_node(
            "LayerNormalization",
            ["residual_out", "ln2_weight", "ln2_bias"],
            ["output"],
            "layer_norm_2",
            axis=-1,
            epsilon=1e-5,
        )
    )

    # ------------------------------------------------------------------
    # Assemble and return the model
    # ------------------------------------------------------------------
    graph = helper.make_graph(
        nodes,
        "bart_sdpa_attention",
        graph_inputs,
        [graph_output],
        initializers,
    )
    opset = helper.make_opsetid("", 18)
    model = helper.make_model(graph, opset_imports=[opset])
    model.ir_version = 9
    return model


if __name__ == "__main__":
    import os

    output_dir = os.path.dirname(__file__)

    model_with_mask = create_bart_attention_sdpa(hidden_size=16, num_heads=4, with_mask=True)
    path_with_mask = os.path.join(output_dir, "bart_sdpa_attention_with_mask.onnx")
    onnx.save(model_with_mask, path_with_mask)
    print(f"Saved: {path_with_mask}")

    model_no_mask = create_bart_attention_sdpa(hidden_size=16, num_heads=4, with_mask=False)
    path_no_mask = os.path.join(output_dir, "bart_sdpa_attention_no_mask.onnx")
    onnx.save(model_no_mask, path_no_mask)
    print(f"Saved: {path_no_mask}")



import argparse
import hashlib
import os
import platform
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path


def clean_gpg_trustdb() -> None:
    if platform.system() == "Windows":
        # Clean up GPG trust database if it exists
        appdata = os.environ.get("APPDATA")
        if appdata:
            trustdb_path = Path(appdata) / "gnupg" / "trustdb.gpg"
            if trustdb_path.exists():
                try:
                    trustdb_path.unlink()
                except OSError as e:
                    print(f"Warning: failed to delete GPG trust database at {trustdb_path}: {e}")


def get_gpg_path() -> Path:
    """Finds the path to the GPG executable."""
    if platform.system() == "Windows":
        # Check both Program Files (x86) and Program Files, gpg distribution
        # changed on 12/31/2025
        program_files_x86 = os.environ.get("ProgramFiles(x86)")  # noqa: SIM112
        program_files = os.environ.get("ProgramFiles")  # noqa: SIM112

        for base_path in [program_files_x86, program_files]:
            if base_path:
                gpg_path = Path(base_path) / "gnupg/bin/gpg.exe"
                if gpg_path.is_file():
                    return gpg_path

        raise FileNotFoundError("GPG executable not found in Program Files or Program Files (x86).")

    gpg_path_str = shutil.which("gpg")
    if gpg_path_str is None:
        raise FileNotFoundError("gpg executable not found in system PATH.")
    return Path(gpg_path_str)


def run_command(command: list[str], check: bool = True) -> subprocess.CompletedProcess:
    """Executes a command and raises an exception if it fails."""
    print(f"Running command: {' '.join(command)}")
    result = subprocess.run(command, capture_output=True, text=True, check=False)
    if check and result.returncode != 0:
        print(f"Command failed with exit code {result.returncode}")
        print(f"Stdout:\n{result.stdout}")
        print(f"Stderr:\n{result.stderr}")
        raise subprocess.CalledProcessError(result.returncode, command, result.stdout, result.stderr)
    return result


def create_hash_file(file_path: Path, algorithm: str) -> None:
    """Creates a checksum file for the given file using the specified algorithm."""
    print(f"  - Generating {algorithm.upper()} checksum...")
    try:
        hasher = hashlib.new(algorithm)
        with file_path.open("rb") as f:
            # Read in chunks to handle large files efficiently
            while chunk := f.read(8192):
                hasher.update(chunk)

        hash_value = hasher.hexdigest()
        # Create checksum file in 'sha1sum'/'md5sum' format.
        # The '*' indicates to read the file in binary mode for verification tools.
        Path(f"{file_path}.{algorithm}").write_text(hash_value.lower(), encoding="utf-8")
    except Exception as e:
        print(f"Error generating {algorithm} hash for {file_path}: {e}")
        raise


def main() -> None:
    """
    Signs files with GPG and generates checksums.
    """
    parser = argparse.ArgumentParser(description="Signs files with GPG and generates checksums.")
    parser.add_argument("jar_file_directory", help="The directory containing files to sign.")
    args = parser.parse_args()

    jar_file_directory = Path(args.jar_file_directory)
    if not jar_file_directory.is_dir():
        print(f"Error: Directory not found at '{jar_file_directory}'", file=sys.stderr)
        sys.exit(1)

    print(f"\nListing files to be processed in '{jar_file_directory}':")
    files_to_process = [p for p in jar_file_directory.rglob("*") if p.is_file()]
    for file_path in files_to_process:
        print(f"  - {file_path}")
    print(f"Found {len(files_to_process)} files.")

    print("\nGetting GnuPG signing keys from environment variables.")
    gpg_passphrase = os.environ.get("JAVA_PGP_PWD")
    gpg_private_key = os.environ.get("JAVA_PGP_KEY")

    if not gpg_passphrase or not gpg_private_key:
        print(
            "Error: GPG passphrase or private key not found in environment variables ('JAVA_PGP_PWD', 'JAVA_PGP_KEY').",
            file=sys.stderr,
        )
        sys.exit(1)

    gpg_exe_path = get_gpg_path()
    if not gpg_exe_path.is_file():
        print(f"Error: GPG executable not found at '{gpg_exe_path}'.", file=sys.stderr)
        sys.exit(1)

    agent_temp_dir = os.environ.get("AGENT_TEMPDIRECTORY")

    # Use a single temporary directory to manage all temporary files
    with tempfile.TemporaryDirectory(dir=agent_temp_dir) as temp_dir:
        temp_dir_path = Path(temp_dir)
        print(f"Created temporary directory: {temp_dir_path}")

        private_key_file = temp_dir_path / "private.key"
        passphrase_file = temp_dir_path / "passphrase.txt"

        print("Writing GnuPG key and passphrase to temporary files.")
        private_key_file.write_text(gpg_private_key, encoding="utf-8")
        passphrase_file.write_text(gpg_passphrase, encoding="utf-8")

        # Clear GPG trustdb to avoid issues with cached trust settings
        clean_gpg_trustdb()

        # Check GPG trust database to generate a fresh one
        print("Checking trust db")
        run_command([str(gpg_exe_path), "--check-trustdb"])

        print("Importing GnuPG private key.")
        run_command([str(gpg_exe_path), "--batch", "--yes", "--import", str(private_key_file)])
        print("Successfully imported GnuPG private key.")

        print(f"\nProcessing {len(files_to_process)} files in '{jar_file_directory}'.")

        for file_path in files_to_process:
            print(f"Processing file: {file_path}")

            # GPG Signing (.asc)
            print("  - GnuPG signing...")
            run_command(
                [
                    str(gpg_exe_path),
                    "--pinentry-mode",
                    "loopback",
                    "--passphrase-file",
                    str(passphrase_file),
                    "--detach-sign",
                    "--armor",
                    str(file_path),
                ]
            )

            # SHA-1 and MD5 Checksums
            create_hash_file(file_path, "sha1")
            create_hash_file(file_path, "md5")

    print("\nFile signing and checksum generation completed.")
    print("Temporary directory and its contents have been deleted.")


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# This code is used to generate test data for the test case:
# TEST(ResizeOpTest, NoAntialias_AlignCorners_Cubic_Floor_NHWC)
# located in onnxruntime/test/providers/cpu/tensor/resize_op_test.cc

import torch
import torch.nn.functional as F

# Define input tensor
input_tensor = (
    torch.tensor(
        [
            [[1, 2], [3, 4], [5, 6], [7, 8]],
            [[9, 10], [11, 12], [13, 14], [15, 16]],
            [[17, 18], [19, 20], [21, 22], [23, 24]],
        ]
    )
    .unsqueeze(0)
    .to(torch.float32)
)

output_shape = (1, 6, 8, 2)

# Apply resize operation
output_tensor = F.interpolate(
    input_tensor.permute(0, 3, 1, 2),  # Convert to NHWC to NCHW for PyTorch
    size=(6, 8),  # Note: PyTorch expects size in (height, width) format
    mode="bicubic",  # bicubic is the same as cubic in ONNX
    align_corners=True,
).permute(0, 2, 3, 1)  # Convert back to NHWC

print(output_tensor)



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft, Intel Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import abc
import contextlib
import copy
import itertools
import json
import os
import tempfile
import uuid
from collections.abc import Sequence
from enum import Enum
from pathlib import Path

import numpy as np
import onnx
from onnx import ModelProto, TensorProto, helper, numpy_helper

import onnxruntime

from .quant_utils import apply_plot, load_model_with_shape_infer, smooth_distribution


def rel_entr(pk: np.ndarray, qk: np.ndarray) -> np.ndarray:
    """
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.rel_entr.html#scipy.special.rel_entr.
    Python implementation.
    """
    res = np.empty(pk.shape, dtype=pk.dtype)
    res[:] = pk[:] * np.log(pk[:] / qk[:])
    c2 = (pk == 0) & (qk >= 0)
    res[c2] = 0
    c1 = (pk > 0) & (qk > 0)
    res[~c1] = np.inf
    return res


def entropy(
    pk: np.ndarray,
    qk: np.ndarray,
    base: float | None = None,
    axis: int = 0,
) -> np.ndarray:
    """
    Simplifeied version of entropy.
    Source: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.entropy.html.
    This avoids taking a dependency on scipy just for this function.
    """
    assert base is None or base > 0, "base={base} must be a positive number or `None`."
    assert qk is not None, "qk is None"

    pk = np.asarray(pk).astype(np.float32)
    pk = 1.0 * pk / np.sum(pk, axis=axis, keepdims=True)

    qk = np.asarray(qk).astype(np.float32)
    pk, qk = np.broadcast_arrays(pk, qk)
    qk = 1.0 * qk / np.sum(qk, axis=axis, keepdims=True)
    vec = rel_entr(pk, qk)

    s = np.sum(vec, axis=axis)
    if base is not None:
        s /= np.log(base)
    return s.astype(pk.dtype)


class TensorData:
    _allowed = frozenset(["avg", "std", "lowest", "highest", "hist", "hist_edges", "bins"])
    _floats = frozenset(["avg", "std", "lowest", "highest", "hist_edges"])

    def __init__(self, **kwargs):
        self._attrs = list(kwargs.keys())
        for k, v in kwargs.items():
            if k not in TensorData._allowed:
                raise ValueError(f"Unexpected value {k!r} not in {TensorData._allowed}.")
            if k in TensorData._floats:
                if not hasattr(v, "dtype"):
                    raise ValueError(f"Unexpected type {type(v)} for k={k!r}")
                if v.dtype not in (np.float16, np.float32):
                    raise ValueError(f"Unexpected dtype {v.dtype} for k={k!r}")
            setattr(self, k, v)

    @property
    def range_value(self):
        if not hasattr(self, "lowest") or not hasattr(self, "highest"):
            raise AttributeError(f"Attributes 'lowest' and/or 'highest' missing in {dir(self)}.")
        return (self.lowest, self.highest)

    @property
    def avg_std(self):
        if not hasattr(self, "avg") or not hasattr(self, "std"):
            raise AttributeError(f"Attributes 'avg' and/or 'std' missing in {dir(self)}.")
        return (self.avg, self.std)

    def to_dict(self):
        # This is needed to serialize the data into JSON.
        data = {k: getattr(self, k) for k in self._attrs}
        data["CLS"] = self.__class__.__name__
        return data

    @classmethod
    def from_dict(cls, d: dict) -> "TensorData":
        """Reconstruct a TensorData from a dict produced by to_dict()."""
        kwargs = {}
        for k, v in d.items():
            if k == "CLS":
                continue
            value = v
            if isinstance(value, dict) and value.get("CLS") == "numpy.array":
                value = np.array(value["data"], dtype=np.dtype(value["dtype"]))
            elif k in cls._floats and isinstance(value, (int, float)):
                value = np.array(value, dtype=np.float32)
            kwargs[k] = value
        return cls(**kwargs)


class TensorsData:
    def __init__(self, calibration_method, data: dict[str, TensorData | tuple]):
        self.calibration_method = calibration_method
        self.data = {}
        for k, v in data.items():
            if not isinstance(k, str):
                raise TypeError(f"Keys must be strings not {type(k)}.")
            if isinstance(v, tuple):
                if calibration_method == CalibrationMethod.MinMax and len(v) == 2:
                    self.data[k] = TensorData(lowest=v[0], highest=v[1])
                    continue
                if len(v) == 4:
                    self.data[k] = TensorData(lowest=v[0], highest=v[1], hist=v[2], bins=v[3])
                    continue
                raise TypeError(f"Unexpected tuple for {k:r}, it has {len(v)} elements: {v}.")
            if not isinstance(v, TensorData):
                raise TypeError(f"Values must be TensorData not {type(v)}.")
            self.data[k] = v

    def __iter__(self):
        yield from self.data

    def __contains__(self, key):
        return key in self.data

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        if key not in self.data:
            raise RuntimeError(f"Only an existing tensor can be modified, {key!r} is not.")
        self.data[key] = value

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    def to_dict(self):
        # This is needed to serialize the data into JSON.
        data = {
            "CLS": self.__class__.__name__,
            "data": self.data,
            "calibration_method": self.calibration_method,
        }
        return data

    @classmethod
    def from_dict(cls, d: dict) -> "TensorsData":
        """Reconstruct a TensorsData from a dict produced by to_dict()."""
        method_val = d["calibration_method"]
        if isinstance(method_val, dict) and method_val.get("CLS") == "CalibrationMethod":
            name = method_val["value"].split(".")[-1]
            method = CalibrationMethod[name]
        else:
            method = method_val
        reconstructed = {k: TensorData.from_dict(v) for k, v in d["data"].items()}
        return cls(method, reconstructed)


class CalibrationMethod(Enum):
    MinMax = 0
    Entropy = 1
    Percentile = 2
    Distribution = 3


class CalibrationDataReader(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, "get_next") and callable(subclass.get_next)) or NotImplemented

    @abc.abstractmethod
    def get_next(self) -> dict:
        """generate the input data dict for ONNXinferenceSession run"""
        raise NotImplementedError

    def __iter__(self):
        return self

    def __next__(self):
        result = self.get_next()
        if result is None:
            raise StopIteration
        return result

    def __len__(self):
        raise NotImplementedError

    def set_range(self, start_index: int, end_index: int):
        raise NotImplementedError


class CalibrationCacheEncoder(json.JSONEncoder):
    """Shared JSON encoder for calibration caches.

    Handles numpy ndarrays and numpy scalar types (integer/floating) so
    calibration JSON output is consistent across ``save_tensors_data`` and
    ``quant_utils.write_calibration_table``.
    """

    def default(self, obj):
        if isinstance(obj, (TensorData, TensorsData)):
            return obj.to_dict()
        if isinstance(obj, np.ndarray):
            return {"data": obj.tolist(), "dtype": str(obj.dtype), "CLS": "numpy.array"}
        if isinstance(obj, CalibrationMethod):
            return {"CLS": obj.__class__.__name__, "value": str(obj)}
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


def save_tensors_data(tensors_data: "TensorsData", path: "str | Path", *, smooth_quant: bool = False) -> None:
    """Serialize calibration tensor ranges to a JSON file at *path*.

    :param smooth_quant: whether the producing run used SmoothQuant.  Stored in
        the cache so a later load can detect a mismatch and recompute.
    """
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    fd, tmp_name = tempfile.mkstemp(dir=path.parent, prefix=".calibcache_", suffix=".tmp")
    try:
        with os.fdopen(fd, "w") as f:
            payload = tensors_data.to_dict()
            payload["smooth_quant"] = smooth_quant
            json.dump(payload, f, cls=CalibrationCacheEncoder)
            f.flush()
        os.replace(tmp_name, path)
    except BaseException:
        with contextlib.suppress(FileNotFoundError):
            os.unlink(tmp_name)
        raise


def load_tensors_data(path: "str | Path") -> "TensorsData":
    """Load calibration tensor ranges from a JSON file written by save_tensors_data()."""
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Calibration cache not found: {path}")
    if not path.is_file():
        raise ValueError(f"Calibration cache path is not a file: {path}")
    with path.open("r") as f:
        d = json.load(f)
    return TensorsData.from_dict(d)


class CalibraterBase:
    def __init__(
        self,
        model_path: str | Path,
        op_types_to_calibrate: Sequence[str] | None = None,
        augmented_model_path="augmented_model.onnx",
        symmetric=False,
        use_external_data_format=False,
        per_channel=False,
    ):
        """
        :param model_path: ONNX model to calibrate. It should be a model file path
        :param op_types_to_calibrate: operator types to calibrate. By default, calibrate all the float32/float16 tensors.
        :param augmented_model_path: save augmented model to this path.
        :param symmetric: make range of tensor symmetric (central point is 0).
        :param use_external_data_format: use external data format to store model which size is >= 2Gb.
        :param per_channel: whether to compute ranges per each channel.
        """
        if isinstance(model_path, str):
            self.model = load_model_with_shape_infer(Path(model_path))
        elif isinstance(model_path, Path):
            self.model = load_model_with_shape_infer(model_path)
        else:
            raise ValueError("model_path should be model path.")

        self.op_types_to_calibrate = op_types_to_calibrate
        self.augmented_model_path = augmented_model_path
        self.symmetric = symmetric
        self.use_external_data_format = use_external_data_format
        self.per_channel = per_channel

        self.augment_model = None
        self.infer_session = None
        self.execution_providers = ["CPUExecutionProvider"]

    def set_execution_providers(self, execution_providers=["CPUExecutionProvider"]):  # noqa: B006
        """
        reset the execution providers to execute the collect_data. It triggers to re-creating inference session.
        """
        self.execution_providers = execution_providers
        self.create_inference_session()

    def create_inference_session(self):
        """
        create an OnnxRuntime InferenceSession.
        """
        sess_options = onnxruntime.SessionOptions()
        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
        self.infer_session = onnxruntime.InferenceSession(
            self.augmented_model_path,
            sess_options=sess_options,
            providers=self.execution_providers,
        )

    def select_tensors_to_calibrate(self, model: ModelProto):
        """
        select input/output tensors of candidate nodes to calibrate.
        returns:
            tensors (set): set of tensor name.
            value_infos (dict): tensor name to value info.
        """
        value_infos = {vi.name: vi for vi in model.graph.value_info}
        value_infos.update({ot.name: ot for ot in model.graph.output})
        value_infos.update({it.name: it for it in model.graph.input})
        initializer = {init.name for init in model.graph.initializer}

        tensors_to_calibrate = set()
        tensor_type_to_calibrate = {TensorProto.FLOAT, TensorProto.FLOAT16}

        for node in model.graph.node:
            if not self.op_types_to_calibrate or node.op_type in self.op_types_to_calibrate:
                for tensor_name in itertools.chain(node.input, node.output):
                    if tensor_name in value_infos:
                        vi = value_infos[tensor_name]
                        if (
                            vi.type.HasField("tensor_type")
                            and (vi.type.tensor_type.elem_type in tensor_type_to_calibrate)
                            and (tensor_name not in initializer)
                        ):
                            tensors_to_calibrate.add(tensor_name)

        return tensors_to_calibrate, value_infos

    def get_augment_model(self):
        """
        return: augmented onnx model. Call after calling augment_graph
        """
        return self.model

    def augment_graph(self):
        """
        abstract method: augment the input model to prepare for collecting data. It will:
            1. augment the model to be able to collect desired statistics data
            2. save augmented model to augmented_model_paths
        """
        raise NotImplementedError

    def collect_data(self, data_reader: CalibrationDataReader):
        """
        abstract method: collect the tensors that will be used for range computation. It can be called multiple times.
        """
        raise NotImplementedError

    def compute_data(self) -> TensorsData:
        """
        abstract method: compute data based on the calibration method stored in TensorsData
        """
        raise NotImplementedError


class MinMaxCalibrater(CalibraterBase):
    def __init__(
        self,
        model_path: str | Path,
        op_types_to_calibrate: Sequence[str] | None = None,
        augmented_model_path="augmented_model.onnx",
        symmetric=False,
        use_external_data_format=False,
        moving_average=False,
        averaging_constant=0.01,
        max_intermediate_outputs=None,
        per_channel=False,
    ):
        """
        :param model_path: ONNX model to calibrate. It is a model path
        :param op_types_to_calibrate: operator types to calibrate. By default, calibrate all the float32/float16 tensors.
        :param augmented_model_path: save augmented model to this path.
        :param symmetric: make range of tensor symmetric (central point is 0).
        :param use_external_data_format: use external data format to store model which size is >= 2Gb
        :param moving_average: compute the moving average of the minimum and maximum values instead of the global minimum and maximum.
        :param averaging_constant: constant smoothing factor to use when computing the moving average.
        :param max_intermediate_outputs: maximum number of intermediate outputs before an intermediate range is computed.
        :param per_channel: whether to compute ranges per each channel.
        """
        super().__init__(
            model_path,
            op_types_to_calibrate=op_types_to_calibrate,
            augmented_model_path=augmented_model_path,
            symmetric=symmetric,
            use_external_data_format=use_external_data_format,
            per_channel=per_channel,
        )
        self.intermediate_outputs = []
        self.calibrate_tensors_range = None
        self.num_model_outputs = len(self.model.graph.output)
        self.model_original_outputs = {output.name for output in self.model.graph.output}
        self.moving_average = moving_average
        if moving_average and (averaging_constant < 0 or averaging_constant > 1):
            raise ValueError("Invalid averaging constant, which should not be < 0 or > 1.")
        self.averaging_constant = averaging_constant
        self.max_intermediate_outputs = max_intermediate_outputs

    def augment_graph(self):
        """
        Adds ReduceMin and ReduceMax nodes to all quantization_candidates op type nodes in
        model and ensures their outputs are stored as part of the graph output
        :return: augmented ONNX model
        """
        tensors, _ = self.select_tensors_to_calibrate(self.model)
        reshape_shape_name = str(uuid.uuid4())
        reshape_shape = numpy_helper.from_array(np.array([-1], dtype=np.int64), reshape_shape_name)
        self.model.graph.initializer.append(reshape_shape)

        def get_op_version(op_type, model):
            for opset_import in model.opset_import:
                if onnx.defs.has(op_type, opset_import.domain):
                    return opset_import.version
            raise RuntimeError(f"Model does not contain a version for '{op_type}'.")

        def insert_nodes(tensor_name, new_nodes):
            index = next(
                (i for i, x in enumerate(self.model.graph.node) if tensor_name in x.input), len(self.model.graph.node)
            )
            for node in new_nodes:
                self.model.graph.node.insert(index, node)
                index += 1

        def add_reduce_min_max(tensor_name, reduce_op_name):
            # When doing ReduceMax/ReduceMin, ORT can't reduce on dim with value of 0 if 'keepdims' is false.
            # To make the code simple, we always let keepdims to be 1.
            keepdims = 1

            # Adding ReduceMin/ReduceMax nodes: ReduceMin/ReduceMax -> Reshape-> (output)
            reduce_output = tensor_name + "_" + reduce_op_name
            intermediate_output = reduce_output + "_Reshape"
            reduce_node = onnx.helper.make_node(
                reduce_op_name, [tensor_name], [intermediate_output], keepdims=keepdims, name=reduce_output
            )

            reshape_node = onnx.helper.make_node(
                "Reshape",
                inputs=[intermediate_output, reshape_shape_name],
                outputs=[reduce_output],
                name=intermediate_output,
            )

            value_infos = {vi.name: vi for vi in self.model.graph.value_info}
            value_infos.update({o.name: o for o in self.model.graph.output})
            value_infos.update({i.name: i for i in self.model.graph.input})
            if tensor_name in value_infos:
                onnx_type = value_infos[tensor_name].type.tensor_type.elem_type
            else:
                raise ValueError(
                    f"Unable to guess tensor type for tensor {tensor_name!r}, "
                    "running shape inference before quantization may resolve this issue."
                )

            # Include axes in reduce_op when per_channel, always keeping axis=1
            if self.per_channel:
                tensor_rank = len(value_infos[tensor_name].type.tensor_type.shape.dim)
                reduced_axes = [0, *range(2, tensor_rank)]
                # Depending on opset version, axes in ReduceMin/ReduceMax are in attribute or inputs
                if get_op_version(reduce_op_name, self.model) < 18:
                    reduce_node.attribute.append(helper.make_attribute("axes", reduced_axes))
                else:
                    reduce_axes_name = str(uuid.uuid4())
                    reduce_axes = numpy_helper.from_array(np.array(reduced_axes, dtype=np.int64), reduce_axes_name)
                    reduce_node.input.append(reduce_axes_name)
                    self.model.graph.initializer.append(reduce_axes)

            insert_nodes(tensor_name, [reduce_node, reshape_node])
            self.model.graph.output.append(helper.make_tensor_value_info(reduce_output, onnx_type, [None]))

        for tensor in tensors:
            add_reduce_min_max(tensor, "ReduceMin")
            add_reduce_min_max(tensor, "ReduceMax")

        onnx.save(
            self.model,
            self.augmented_model_path,
            save_as_external_data=self.use_external_data_format,
        )

    def clear_collected_data(self):
        self.intermediate_outputs = []

    def collect_data(self, data_reader: CalibrationDataReader):
        while True:
            inputs = data_reader.get_next()
            if not inputs:
                break
            self.intermediate_outputs.append(
                [
                    value if sess_o.name not in self.model_original_outputs else None
                    for sess_o, value in zip(
                        self.infer_session.get_outputs(), self.infer_session.run(None, inputs), strict=False
                    )
                ]
            )
            if (
                self.max_intermediate_outputs is not None
                and len(self.intermediate_outputs) == self.max_intermediate_outputs
            ):
                self.clear_collected_data()

        if len(self.intermediate_outputs) == 0 and self.calibrate_tensors_range is None:
            raise ValueError("No data is collected.")

        t = self.compute_data()
        if not isinstance(t, TensorsData):
            raise TypeError(f"compute_data must return a TensorsData not {type(t)}.")
        self.clear_collected_data()

    def merge_range(self, old_range, new_range):
        if not old_range:
            return new_range

        for key, value in old_range.items():
            # Handling for structured data types with TensorData
            if isinstance(value, TensorData):
                old_min = value.range_value[0]
                old_max = value.range_value[1]
            else:
                old_min, old_max = value

            if isinstance(new_range[key], TensorData):
                new_min = new_range[key].range_value[0]
                new_max = new_range[key].range_value[1]
            else:
                new_min, new_max = new_range[key]

            if self.moving_average:
                min_value = old_min + self.averaging_constant * (new_min - old_min)
                max_value = old_max + self.averaging_constant * (new_max - old_max)
            else:
                min_value = min(old_min, new_min)
                max_value = max(old_max, new_max)

            # If structured as TensorData, wrap the result accordingly
            if isinstance(value, TensorData) or isinstance(new_range[key], TensorData):
                new_range[key] = TensorData(lowest=min_value, highest=max_value)
            else:
                new_range[key] = (min_value, max_value)

        return new_range

    def compute_data(self) -> TensorsData:
        """
        Compute the min-max range of tensor
        :return: dictionary mapping: {added node names: (ReduceMin, ReduceMax) pairs }
        """

        if len(self.intermediate_outputs) == 0:
            return self.calibrate_tensors_range

        output_names = [self.infer_session.get_outputs()[i].name for i in range(len(self.intermediate_outputs[0]))]
        output_dicts_list = [
            dict(zip(output_names, intermediate_output, strict=False))
            for intermediate_output in self.intermediate_outputs
        ]

        merged_output_dict = {}
        for d in output_dicts_list:
            for k, v in d.items():
                merged_output_dict.setdefault(k, []).append(v)
        added_output_names = output_names[self.num_model_outputs :]
        calibrate_tensor_names = [
            added_output_names[i].rpartition("_")[0] for i in range(0, len(added_output_names), 2)
        ]  # output names

        merged_added_output_dict = {
            i: merged_output_dict[i] for i in merged_output_dict if i not in self.model_original_outputs
        }

        pairs = []
        for i in range(0, len(added_output_names), 2):
            if self.moving_average:
                min_value_array = np.nanmean(merged_added_output_dict[added_output_names[i]], axis=0)
                max_value_array = np.nanmean(merged_added_output_dict[added_output_names[i + 1]], axis=0)
            else:
                min_value_array = np.nanmin(merged_added_output_dict[added_output_names[i]], axis=0)
                max_value_array = np.nanmax(merged_added_output_dict[added_output_names[i + 1]], axis=0)

            if self.symmetric:
                max_absolute_value = np.nanmax([np.abs(min_value_array), np.abs(max_value_array)], axis=0)
                pairs.append((-max_absolute_value, max_absolute_value))
            else:
                pairs.append((min_value_array, max_value_array))

        new_calibrate_tensors_range = TensorsData(
            CalibrationMethod.MinMax, dict(zip(calibrate_tensor_names, pairs, strict=False))
        )
        if self.calibrate_tensors_range:
            self.calibrate_tensors_range = self.merge_range(self.calibrate_tensors_range, new_calibrate_tensors_range)
        else:
            self.calibrate_tensors_range = new_calibrate_tensors_range

        return self.calibrate_tensors_range


class HistogramCalibrater(CalibraterBase):
    def __init__(
        self,
        model_path: str | Path,
        op_types_to_calibrate: Sequence[str] | None = None,
        augmented_model_path="augmented_model.onnx",
        use_external_data_format=False,
        method="percentile",
        symmetric=False,
        num_bins=128,
        num_quantized_bins=2048,
        percentile=99.999,
        scenario="same",
    ):
        """
        :param model_path: ONNX model to calibrate. It is a model path.
        :param op_types_to_calibrate: operator types to calibrate. By default, calibrate all the float32/float16 tensors.
        :param augmented_model_path: save augmented model to this path.
        :param use_external_data_format: use external data format to store model which size is >= 2Gb
        :param method: A string. One of ['entropy', 'percentile'].
        :param symmetric: make range of tensor symmetric (central point is 0).
        :param num_bins: number of bins to create a new histogram for collecting tensor values.
        :param num_quantized_bins: number of quantized bins. Default 128.
        :param percentile: A float number between [0, 100]. Default 99.99.
        :param scenario: see :class:`DistributionCalibrater`
        """
        super().__init__(
            model_path,
            op_types_to_calibrate=op_types_to_calibrate,
            augmented_model_path=augmented_model_path,
            symmetric=symmetric,
            use_external_data_format=use_external_data_format,
        )
        self.intermediate_outputs = []
        self.calibrate_tensors_range = None
        self.num_model_outputs = len(self.model.graph.output)
        self.model_original_outputs = {output.name for output in self.model.graph.output}
        self.collector = None
        self.method = method
        self.num_bins = num_bins
        self.num_quantized_bins = num_quantized_bins
        self.percentile = percentile
        self.tensors_to_calibrate = None
        self.scenario = scenario

    def augment_graph(self):
        """
        make all quantization_candidates op type nodes as part of the graph output.
        :return: augmented ONNX model
        """
        self.tensors_to_calibrate, value_infos = self.select_tensors_to_calibrate(self.model)
        for tensor in self.tensors_to_calibrate:
            if tensor not in self.model_original_outputs:
                self.model.graph.output.append(value_infos[tensor])

        onnx.save(
            self.model,
            self.augmented_model_path,
            save_as_external_data=self.use_external_data_format,
        )

    def clear_collected_data(self):
        self.intermediate_outputs = []

    def collect_data(self, data_reader: CalibrationDataReader):
        """
        Entropy Calibrator collects operators' tensors as well as generates tensor histogram for each operator.
        """
        input_names_set = {node_arg.name for node_arg in self.infer_session.get_inputs()}
        output_names = [node_arg.name for node_arg in self.infer_session.get_outputs()]

        while True:
            inputs = data_reader.get_next()
            if not inputs:
                break
            outputs = self.infer_session.run(None, inputs)

            # Copy np.ndarray only for graph outputs that are also graph inputs to workaround bug:
            # https://github.com/microsoft/onnxruntime/issues/21922
            fixed_outputs = []
            for output_index, output in enumerate(outputs):
                if output_names[output_index] in input_names_set:
                    fixed_outputs.append(copy.copy(output))
                else:
                    fixed_outputs.append(output)

            self.intermediate_outputs.append(fixed_outputs)

        if len(self.intermediate_outputs) == 0:
            raise ValueError("No data is collected.")

        output_dicts_list = [
            dict(zip(output_names, intermediate_output, strict=False))
            for intermediate_output in self.intermediate_outputs
        ]

        merged_dict = {}
        for d in output_dicts_list:
            for k, v in d.items():
                merged_dict.setdefault(k, []).append(v)

        clean_merged_dict = {i: merged_dict[i] for i in merged_dict if i in self.tensors_to_calibrate}

        if not self.collector:
            self.collector = HistogramCollector(
                method=self.method,
                symmetric=self.symmetric,
                num_bins=self.num_bins,
                num_quantized_bins=self.num_quantized_bins,
                percentile=self.percentile,
                scenario=self.scenario,
            )
        self.collector.collect(clean_merged_dict)

        self.clear_collected_data()

    def compute_data(self) -> TensorsData:
        """
        Compute the min-max range of tensor
        :return: dictionary mapping: {tensor name: (min value, max value)}
        """
        if not self.collector:
            raise ValueError("No collector created and can't generate calibration data.")

        if isinstance(self, EntropyCalibrater):
            cal = CalibrationMethod.Entropy
        elif isinstance(self, PercentileCalibrater):
            cal = CalibrationMethod.Percentile
        elif isinstance(self, DistributionCalibrater):
            cal = CalibrationMethod.Distribution
        else:
            raise TypeError(f"Unknown calibrater {type(self)}. This method must be overwritten.")
        return TensorsData(cal, self.collector.compute_collection_result())


class EntropyCalibrater(HistogramCalibrater):
    def __init__(
        self,
        model_path: str | Path,
        op_types_to_calibrate: Sequence[str] | None = None,
        augmented_model_path="augmented_model.onnx",
        use_external_data_format=False,
        method="entropy",
        symmetric=False,
        num_bins=128,
        num_quantized_bins=128,
    ):
        """
        :param model_path: ONNX model to calibrate. It is a model path
        :param op_types_to_calibrate: operator types to calibrate. By default, calibrate all the float32/float16 tensors.
        :param augmented_model_path: save augmented model to this path.
        :param use_external_data_format: use external data format to store model which size is >= 2Gb
        :param method: A string. One of ['entropy', 'percentile', 'distribution'].
        :param symmetric: make range of tensor symmetric (central point is 0).
        :param num_bins: number of bins to create a new histogram for collecting tensor values.
        :param num_quantized_bins: number of quantized bins. Default 128.
        """
        super().__init__(
            model_path,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format,
            method=method,
            symmetric=symmetric,
            num_bins=num_bins,
            num_quantized_bins=num_quantized_bins,
        )


class PercentileCalibrater(HistogramCalibrater):
    def __init__(
        self,
        model_path: str | Path,
        op_types_to_calibrate: Sequence[str] | None = None,
        augmented_model_path="augmented_model.onnx",
        use_external_data_format=False,
        method="percentile",
        symmetric=False,
        num_bins=2048,
        percentile=99.999,
    ):
        """
        :param model_path: ONNX model to calibrate. It is a model path
        :param op_types_to_calibrate: operator types to calibrate. By default, calibrate all the float32/float16 tensors.
        :param augmented_model_path: save augmented model to this path.
        :param use_external_data_format: use external data format to store model which size is >= 2Gb
        :param method: A string. One of ['entropy', 'percentile', 'distribution'].
        :param symmetric: make range of tensor symmetric (central point is 0).
        :param num_quantized_bins: number of quantized bins. Default 128.
        :param percentile: A float number between [0, 100]. Default 99.99.
        """
        super().__init__(
            model_path,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format,
            method=method,
            symmetric=symmetric,
            num_bins=num_bins,
            percentile=percentile,
        )


class DistributionCalibrater(HistogramCalibrater):
    def __init__(
        self,
        model_path: str | Path,
        op_types_to_calibrate: Sequence[str] | None = None,
        augmented_model_path="augmented_model.onnx",
        use_external_data_format=False,
        method="distribution",
        num_bins=128,
        scenario="same",
    ):
        """
        :param model_path: ONNX model to calibrate. It is a model path
        :param op_types_to_calibrate: operator types to calibrate. By default, calibrate all the float32/float16 tensors.
        :param augmented_model_path: save augmented model to this path.
        :param use_external_data_format: use external data format to store model which size is >= 2Gb
        :param method: A string. One of ['entropy', 'percentile', 'distribution'].
        :param symmetric: make range of tensor symmetric (central point is 0).
        :param num_bins: number of bins to create a new histogram for collecting tensor values.
        :param scenario: for float 8 only, if `scenario="same"`,
            the algorithm weights and float 8 follow the same distribution,
            if `scenario="p3"`, it assumes the weights follow
            a gaussian law and float 8 ~ X^3 where X is a gaussian law
        """
        super().__init__(
            model_path,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format,
            method=method,
            num_bins=num_bins,
            scenario=scenario,
        )


class CalibrationDataCollector(metaclass=abc.ABCMeta):
    """
    Base class for collecting data for calibration-based quantization.
    """

    @abc.abstractmethod
    def collect(self, name_to_arr):
        """
        Generate informative data based on given data.
            name_to_arr : dict
                tensor name to NDArray data
        """
        raise NotImplementedError

    @abc.abstractmethod
    def compute_collection_result(self):
        """
        Get the optimal result among collection data.
        """
        raise NotImplementedError


class HistogramCollector(CalibrationDataCollector):
    """
    Collecting histogram for each tensor. Percentile and Entropy method are supported.

    ref: https://github.com//apache/incubator-mxnet/blob/master/python/mxnet/contrib/quantization.py
    ref: https://docs.nvidia.com/deeplearning/tensorrt/pytorch-quantization-toolkit/docs/_modules/
                 pytorch_quantization/calib/histogram.html
    """

    def __init__(self, method, symmetric, num_bins, num_quantized_bins, percentile, scenario):
        self.histogram_dict = {}
        self.method = method
        self.symmetric = symmetric
        self.num_bins = num_bins
        self.num_quantized_bins = num_quantized_bins
        self.percentile = percentile
        self.scenario = scenario

    def get_histogram_dict(self):
        return self.histogram_dict

    def collect(self, name_to_arr):
        print("Collecting tensor data and making histogram ...")

        # TODO: Currently we have different collect() for entropy and percentile method respectively.
        #       Need unified collect in the future.
        if self.method in {"distribution", "entropy"}:
            return self.collect_value(name_to_arr)
        elif self.method == "percentile":
            if self.symmetric:
                return self.collect_absolute_value(name_to_arr)
            else:
                return self.collect_value(name_to_arr)
        else:
            raise ValueError("Only 'entropy', 'percentile' or 'distribution' methods are supported")

    def collect_absolute_value(self, name_to_arr):
        """
        Collect histogram on absolute value
        """
        for tensor, data_arr in name_to_arr.items():
            if isinstance(data_arr, list):
                for arr in data_arr:
                    assert isinstance(arr, np.ndarray), f"Unexpected type {type(arr)} for tensor={tensor!r}"
                dtypes = {a.dtype for a in data_arr}
                assert len(dtypes) == 1, (
                    f"The calibration expects only one element type but got {dtypes} for tensor={tensor!r}"
                )
                data_arr_np = np.asarray(data_arr)
            elif not isinstance(data_arr, np.ndarray):
                raise ValueError(f"Unexpected type {type(data_arr)} for tensor={tensor!r}")
            else:
                data_arr_np = data_arr
            data_arr_np = data_arr_np.flatten()
            if data_arr_np.size > 0:
                min_value = np.nanmin(data_arr_np)
                max_value = np.nanmax(data_arr_np)
            else:
                min_value = np.array(0, dtype=data_arr_np.dtype)
                max_value = np.array(0, dtype=data_arr_np.dtype)

            data_arr_np = np.absolute(data_arr_np)  # only consider absolute value

            if tensor not in self.histogram_dict:
                # first time it uses num_bins to compute histogram.
                hist, hist_edges = np.histogram(data_arr_np, bins=self.num_bins)
                hist_edges = hist_edges.astype(data_arr_np.dtype)
                assert data_arr_np.dtype != np.float64, (
                    "only float32 or float16 is supported, every constant must be explicitly typed"
                )
                self.histogram_dict[tensor] = (hist, hist_edges, min_value, max_value)
            else:
                old_histogram = self.histogram_dict[tensor]
                old_min = old_histogram[2]
                old_max = old_histogram[3]
                assert hasattr(old_min, "dtype"), f"old_min should be a numpy array but is {type(old_min)}"
                assert hasattr(old_max, "dtype"), f"old_min should be a numpy array but is {type(old_max)}"
                old_hist = old_histogram[0]
                old_hist_edges = old_histogram[1]
                temp_amax = np.nanmax(data_arr_np)
                if temp_amax > old_hist_edges[-1]:
                    # increase the number of bins
                    width = old_hist_edges[1] - old_hist_edges[0]
                    # NOTE: np.arange may create an extra bin after the one containing temp_amax
                    new_bin_edges = np.arange(old_hist_edges[-1] + width, temp_amax + width, width)
                    old_hist_edges = np.hstack((old_hist_edges, new_bin_edges))
                hist, hist_edges = np.histogram(data_arr_np, bins=old_hist_edges)
                hist_edges = hist_edges.astype(data_arr_np.dtype)
                hist[: len(old_hist)] += old_hist
                assert data_arr_np.dtype != np.float64, (
                    "only float32 or float16 is supported, every constant must be explicitly typed"
                )
                self.histogram_dict[tensor] = (hist, hist_edges, min(old_min, min_value), max(old_max, max_value))

    def collect_value(self, name_to_arr):
        """
        Collect histogram on real value
        """
        for tensor, data_arr in name_to_arr.items():
            data_arr = np.asarray(data_arr)  # noqa: PLW2901
            data_arr = data_arr.flatten()  # noqa: PLW2901

            if data_arr.size > 0:
                min_value = np.nanmin(data_arr)
                max_value = np.nanmax(data_arr)
            else:
                min_value = np.array(0, dtype=data_arr.dtype)
                max_value = np.array(0, dtype=data_arr.dtype)

            threshold = np.array(max(abs(min_value), abs(max_value)), dtype=data_arr.dtype)

            if tensor in self.histogram_dict:
                old_histogram = self.histogram_dict[tensor]
                self.histogram_dict[tensor] = self.merge_histogram(
                    old_histogram, data_arr, min_value, max_value, threshold
                )
            else:
                hist, hist_edges = np.histogram(data_arr, self.num_bins, range=(-threshold, threshold))
                self.histogram_dict[tensor] = (
                    hist,
                    hist_edges,
                    min_value,
                    max_value,
                    threshold,
                )

    def merge_histogram(self, old_histogram, data_arr, new_min, new_max, new_threshold):
        (old_hist, old_hist_edges, old_min, old_max, old_threshold) = old_histogram

        if new_threshold <= old_threshold:
            new_hist, _ = np.histogram(data_arr, len(old_hist), range=(-old_threshold, old_threshold))
            return (
                new_hist + old_hist,
                old_hist_edges,
                min(old_min, new_min),
                max(old_max, new_max),
                old_threshold,
            )
        else:
            if old_threshold == 0:
                hist, hist_edges = np.histogram(data_arr, len(old_hist), range=(-new_threshold, new_threshold))
                hist += old_hist
            else:
                old_num_bins = len(old_hist)
                old_stride = 2 * old_threshold / old_num_bins
                half_increased_bins = int((new_threshold - old_threshold) // old_stride + 1)
                new_num_bins = old_num_bins + 2 * half_increased_bins
                new_threshold = half_increased_bins * old_stride + old_threshold
                hist, hist_edges = np.histogram(data_arr, new_num_bins, range=(-new_threshold, new_threshold))
                hist[half_increased_bins : new_num_bins - half_increased_bins] += old_hist
            return (
                hist,
                hist_edges,
                min(old_min, new_min),
                max(old_max, new_max),
                new_threshold,
            )

    def compute_collection_result(self):
        if not self.histogram_dict or len(self.histogram_dict) == 0:
            raise ValueError("Histogram has not been collected. Please run collect() first.")
        print(f"Finding optimal threshold for each tensor using {self.method!r} algorithm ...")

        if self.method == "entropy":
            return self.compute_entropy()
        elif self.method == "percentile":
            return self.compute_percentile()
        elif self.method == "distribution":
            return self.compute_distribution()
        else:
            raise ValueError("Only 'entropy', 'percentile' or 'distribution' methods are supported")

    def compute_percentile(self):
        if self.percentile < 0 or self.percentile > 100:
            raise ValueError("Invalid percentile. Must be in range 0 <= percentile <= 100.")

        histogram_dict = self.histogram_dict
        percentile = self.percentile

        thresholds_dict = {}  # per tensor thresholds

        print(f"Number of tensors : {len(histogram_dict)}")
        print(f"Number of histogram bins : {self.num_bins}")
        print(f"Percentile : ({100.0 - percentile},{percentile})")

        for tensor, histogram in histogram_dict.items():
            hist = histogram[0]
            hist_edges = histogram[1]
            total = hist.sum()
            cdf = np.cumsum(hist / total)
            if self.symmetric:
                idx_right = np.searchsorted(cdf, percentile / 100.0)

                thresholds_dict[tensor] = (
                    -np.array(hist_edges[idx_right], dtype=hist_edges.dtype),
                    np.array(hist_edges[idx_right], dtype=hist_edges.dtype),
                )
            else:
                percent_to_cut_one_side = (100.0 - percentile) / 200.0
                idx_right = np.searchsorted(cdf, 1.0 - percent_to_cut_one_side)
                idx_left = np.searchsorted(cdf, percent_to_cut_one_side)
                thresholds_dict[tensor] = (
                    np.array(hist_edges[idx_left], dtype=hist_edges.dtype),
                    np.array(hist_edges[idx_right], dtype=hist_edges.dtype),
                )
            min_value = histogram[2]
            max_value = histogram[3]
            if thresholds_dict[tensor][0] < min_value:
                thresholds_dict[tensor] = (min_value, thresholds_dict[tensor][1])
            if thresholds_dict[tensor][1] > max_value:
                thresholds_dict[tensor] = (thresholds_dict[tensor][0], max_value)
            thresholds_dict[tensor] = (*thresholds_dict[tensor], *hist[:2])
            # Plot histogram for debug only
            if os.environ.get("QUANTIZATION_DEBUG", "0") in (1, "1"):
                apply_plot(hist, hist_edges)

        return thresholds_dict

    def compute_entropy(self):
        histogram_dict = self.histogram_dict
        num_quantized_bins = self.num_quantized_bins

        thresholds_dict = {}  # per tensor thresholds

        print(f"Number of tensors : {len(histogram_dict)}")
        print(f"Number of histogram bins : {self.num_bins} (The number may increase depends on the data it collects)")
        print(f"Number of quantized bins : {self.num_quantized_bins}")

        for tensor, histogram in histogram_dict.items():
            optimal_threshold = self.get_entropy_threshold(histogram, num_quantized_bins)
            thresholds_dict[tensor] = optimal_threshold
            thresholds_dict[tensor] = (*optimal_threshold, *histogram[:2])

            # Plot histogram for debug only
            if os.environ.get("QUANTIZATION_DEBUG", "0") in (1, "1"):
                apply_plot(histogram[0], histogram[1])

        return thresholds_dict

    @staticmethod
    def _avg_std(hist, hist_edges, power=1):
        if power <= 0:
            raise ValueError(f"power={power} <= 0 is invalid.")
        values = (hist_edges[:-1] + hist_edges[1:]) * 0.5
        if power == 1:
            avg = (hist * values).sum() / hist.sum()
            std = ((hist * values**2).sum() / hist.sum() - avg**2) ** 0.5
            return np.array(avg, dtype=hist_edges.dtype), np.array(std, dtype=hist_edges.dtype)
        if int(power) == power and int(power) % 2 == 1:
            avg = (hist * values**power).sum() / hist.sum()
            std = ((hist * (values**power - avg) ** 2).sum() / hist.sum()) ** 0.5
            return np.array(avg, dtype=hist_edges.dtype), np.array(std, dtype=hist_edges.dtype)

        fact = np.abs(values) / values
        fact[np.isnan(fact)] = 1
        fact[np.isinf(fact)] = 1
        values = np.abs(values) ** power * fact
        avg = (hist * values).sum() / hist.sum()
        std = ((hist * values**2).sum() / hist.sum() - avg**2) ** 0.5
        return np.array(avg, dtype=hist_edges.dtype), np.array(std, dtype=hist_edges.dtype)

    def compute_distribution(self):
        if self.num_bins < 512:
            raise ValueError("Invalid num_bins. Must be in range 512 <= num_bins.")

        histogram_dict = self.histogram_dict
        thresholds_dict = {}  # per tensor thresholds

        print(f"Number of tensors : {len(histogram_dict)}")
        print(f"Number of histogram bins : {self.num_bins}")
        print(f"Scenario : {self.scenario!r})")

        for tensor, histogram in histogram_dict.items():
            hist = histogram[0]
            hist_edges = histogram[1]

            assert hist_edges.dtype != np.float64
            if self.scenario == "same":
                avg_coef, std_coef = self._avg_std(hist, hist_edges, power=1)
            elif self.scenario == "p3":
                avg_coef, std_coef = self._avg_std(hist, hist_edges, power=1.0 / 3.0)
            else:
                raise ValueError("Invalid scenario. Must be in {'same', 'p3'}.")
            assert avg_coef.dtype != np.float64
            assert std_coef.dtype != np.float64
            assert hist_edges.dtype != np.float64
            thresholds_dict[tensor] = TensorData(
                avg=avg_coef,
                std=std_coef,
                hist=hist,
                hist_edges=hist_edges,
                lowest=hist_edges.min(),
                highest=hist_edges.max(),
            )

            # Plot histogram for debug only
            if os.environ.get("QUANTIZATION_DEBUG", "0") in (1, "1"):
                apply_plot(hist, hist_edges)

        return thresholds_dict

    def get_entropy_threshold(self, histogram, num_quantized_bins):
        """Given a dataset, find the optimal threshold for quantizing it.
        The reference distribution is `q`, and the candidate distribution is `p`.
        `q` is a truncated version of the original distribution.
        Ref: http://on-demand.gputechconf.com/gtc/2017/presentation/s7310-8-bit-inference-with-tensorrt.pdf
        """
        hist = histogram[0]
        hist_edges = histogram[1]
        num_bins = hist.size
        zero_bin_index = num_bins // 2
        num_half_quantized_bin = num_quantized_bins // 2

        dtype = histogram[1].dtype
        kl_divergence = np.zeros(zero_bin_index - num_half_quantized_bin + 1)
        thresholds = [(np.array(0, dtype=dtype), np.array(0, dtype=dtype)) for i in range(kl_divergence.size)]

        # <------------ num bins ---------------->
        #        <--- quantized bins ---->
        # |======|===========|===========|=======|
        #              zero bin index
        #        ^                       ^
        #        |                       |
        #   start index               end index          (start of iteration)
        #     ^                             ^
        #     |                             |
        #  start index                  end index               ...
        # ^                                      ^
        # |                                      |
        # start index                    end index       (end of iteration)

        for i in range(num_half_quantized_bin, zero_bin_index + 1, 1):
            start_index = zero_bin_index - i
            end_index = min(zero_bin_index + i + 1, num_bins)

            thresholds[i - num_half_quantized_bin] = (hist_edges[start_index], hist_edges[end_index])

            sliced_distribution = copy.deepcopy(hist[start_index:end_index])

            # reference distribution p
            p = sliced_distribution.copy()  # a copy of np array
            left_outliers_count = sum(hist[:start_index])
            right_outliers_count = sum(hist[end_index:])
            p[0] += left_outliers_count
            p[-1] += right_outliers_count

            # nonzeros[i] incidates whether p[i] is non-zero
            nonzeros = (p != 0).astype(np.int64)

            # quantize p.size bins into quantized bins (default 128 bins)
            quantized_bins = np.zeros(num_quantized_bins, dtype=np.int64)
            num_merged_bins = sliced_distribution.size // num_quantized_bins

            # merge bins into quantized bins
            for index in range(num_quantized_bins):
                start = index * num_merged_bins
                end = start + num_merged_bins
                quantized_bins[index] = sum(sliced_distribution[start:end])
            quantized_bins[-1] += sum(sliced_distribution[num_quantized_bins * num_merged_bins :])

            # in order to compare p and q, we need to make length of q equals to length of p
            # expand quantized bins into p.size bins
            q = np.zeros(p.size, dtype=np.int64)
            for index in range(num_quantized_bins):
                start = index * num_merged_bins
                end = start + num_merged_bins

                norm = sum(nonzeros[start:end])
                if norm != 0:
                    q[start:end] = quantized_bins[index] / norm

            p = smooth_distribution(p)
            q = smooth_distribution(q)
            if p is None or q is None:
                div = np.array(np.inf, dtype=dtype)
            else:
                div = np.array(entropy(p, q), dtype=dtype)
            kl_divergence[i - num_half_quantized_bin] = div

        min_kl_divergence_idx = np.argmin(kl_divergence)
        optimal_threshold = thresholds[min_kl_divergence_idx]
        min_value = histogram[2]
        max_value = histogram[3]
        if optimal_threshold[0] < min_value:
            optimal_threshold = (min_value, optimal_threshold[1])
        if optimal_threshold[1] > max_value:
            optimal_threshold = (optimal_threshold[0], max_value)
        assert hasattr(optimal_threshold[0], "dtype")
        assert hasattr(optimal_threshold[1], "dtype")
        return optimal_threshold


def create_calibrator(
    model: str | Path,
    op_types_to_calibrate: Sequence[str] | None = None,
    augmented_model_path="augmented_model.onnx",
    calibrate_method=CalibrationMethod.MinMax,
    use_external_data_format=False,
    providers=None,
    extra_options={},  # noqa: B006
):
    calibrator = None
    if calibrate_method == CalibrationMethod.MinMax:
        # default settings for min-max algorithm
        symmetric = extra_options.get("symmetric", False)
        moving_average = extra_options.get("moving_average", False)
        averaging_constant = extra_options.get("averaging_constant", 0.01)
        max_intermediate_outputs = extra_options.get("max_intermediate_outputs", None)
        per_channel = extra_options.get("per_channel", False)
        calibrator = MinMaxCalibrater(
            model,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format=use_external_data_format,
            symmetric=symmetric,
            moving_average=moving_average,
            averaging_constant=averaging_constant,
            max_intermediate_outputs=max_intermediate_outputs,
            per_channel=per_channel,
        )
    elif calibrate_method == CalibrationMethod.Entropy:
        # default settings for entropy algorithm
        num_bins = extra_options.get("num_bins", 128)
        num_quantized_bins = extra_options.get("num_quantized_bins", 128)
        symmetric = extra_options.get("symmetric", False)
        calibrator = EntropyCalibrater(
            model,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format=use_external_data_format,
            symmetric=symmetric,
            num_bins=num_bins,
            num_quantized_bins=num_quantized_bins,
        )
    elif calibrate_method == CalibrationMethod.Percentile:
        # default settings for percentile algorithm
        num_bins = extra_options.get("num_bins", 2048)
        percentile = extra_options.get("percentile", 99.999)
        symmetric = extra_options.get("symmetric", True)
        calibrator = PercentileCalibrater(
            model,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format=use_external_data_format,
            symmetric=symmetric,
            num_bins=num_bins,
            percentile=percentile,
        )

    elif calibrate_method == CalibrationMethod.Distribution:
        # default settings for percentile algorithm
        num_bins = extra_options.get("num_bins", 2048)
        scenario = extra_options.get("scenario", "same")

        calibrator = DistributionCalibrater(
            model,
            op_types_to_calibrate,
            augmented_model_path,
            use_external_data_format=use_external_data_format,
            num_bins=num_bins,
            scenario=scenario,
        )

    if calibrator:
        calibrator.augment_graph()
        if providers:
            calibrator.execution_providers = providers
        calibrator.create_inference_session()
        return calibrator

    raise ValueError(f"Unsupported calibration method {calibrate_method}")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""
Cherry-Pick Helper Script
-------------------------
Description:
    This script automates the process of cherry-picking commits for a release branch.
    It fetches merged PRs with a specific label, sorts them by merge date, and generates:
    1. A batch file (.cmd) with git cherry-pick commands.
    2. A markdown file (.md) for the PR description.
    It also checks for potential missing dependencies (conflicts) by verifying if files modified
    by the cherry-picked commits have any other modifications in commits that are not in the
    specified target branch and are not included in the cherry-pick list.

Usage:
    python cherry_pick.py --label "release:1.24.2" --output cherry_pick.cmd --branch "origin/rel-1.24.2"

Requirements:
    - Python 3.7+
    - GitHub CLI (gh) logged in.
    - Git available in PATH.
"""

import argparse
import json
import os
import sys

from cherry_pick_utils import (
    check_preflight,
    extract_pr_numbers,
    get_pr_number_from_subject,
    run_command,
)


def get_merged_prs(repo, label, limit=200):
    """Fetch merged PRs with the specific label."""
    print(f"Fetching merged PRs with label '{label}' from {repo}...")
    cmd = [
        "gh",
        "pr",
        "list",
        "--repo",
        repo,
        "--label",
        label,
        "--state",
        "merged",
        "--json",
        "number,title,mergeCommit,mergedAt",
        "-L",
        str(limit),
    ]
    output = run_command(cmd)
    if not output:
        return []

    try:
        return json.loads(output)
    except json.JSONDecodeError as e:
        print(f"Error parsing gh output: {e}", file=sys.stderr)
        return []


def get_changed_files(oid):
    """Get list of files changed in a commit."""
    output = run_command(["git", "diff-tree", "--no-commit-id", "--name-only", "-m", "-r", oid], silent=True)
    if output:
        return output.strip().splitlines()
    return []


def sanitize_title(title):
    """Normalize PR titles for single-line text output."""
    return title.replace("\n", " ").strip()


def escape_markdown_table_cell(text):
    """Escape markdown table delimiters in cell content."""
    return sanitize_title(text).replace("|", "\\|")


def get_existing_pr_numbers(branch, repo=None, log_limit=500):
    """Get the set of PR numbers already present in the target branch."""
    output = run_command(["git", "log", branch, "--oneline", "-n", str(log_limit)], silent=True)
    if not output:
        return set()
    pr_numbers = set()

    # Pre-fetch PR cache to avoid redundant gh calls
    pr_cache = {}

    # Process commit log
    lines = output.strip().splitlines()
    for line in lines:
        parts = line.split(" ", 1)
        if len(parts) < 2:
            continue
        subject = parts[1]

        pr_num = get_pr_number_from_subject(subject)
        if not pr_num:
            continue

        pr_num_int = int(pr_num)
        pr_numbers.add(pr_num_int)

        # Check if it's a cherry-pick / meta-PR
        is_meta_pr = (
            "cherry pick" in subject.lower() or "cherry-pick" in subject.lower() or "cherrypick" in subject.lower()
        )

        if is_meta_pr:
            # Query gh to get more details (body/commits) to find squashed sub-PRs
            if pr_num not in pr_cache:
                gh_cmd = ["gh", "pr", "view", pr_num, "--json", "title,body,commits"]
                if repo:
                    gh_cmd.extend(["--repo", repo])
                gh_out = run_command(gh_cmd, silent=True)
                if gh_out:
                    try:
                        pr_cache[pr_num] = json.loads(gh_out)
                    except json.JSONDecodeError:
                        pr_cache[pr_num] = None
                else:
                    pr_cache[pr_num] = None

            details = pr_cache.get(pr_num)
            if details:
                # Collect sub-PRs from title, body, and commits
                extracted_nums = []
                extracted_nums.extend(extract_pr_numbers(details.get("title", ""), strict=True))
                extracted_nums.extend(extract_pr_numbers(details.get("body", ""), strict=True))

                for commit in details.get("commits", []):
                    extracted_nums.extend(extract_pr_numbers(commit.get("messageHeadline", ""), strict=True))

                for num in set(extracted_nums):
                    if num != pr_num_int:
                        pr_numbers.add(num)

    return pr_numbers


def check_missing_dependencies(prs, branch):
    """Check for potential missing dependencies (conflicts)."""
    print("\nChecking for potential missing dependencies (conflicts)...")

    # Collect OIDs being cherry-picked and all their ancestor commits
    cherry_pick_oids = set()
    for pr in prs:
        if pr.get("mergeCommit"):
            merge_oid = pr["mergeCommit"]["oid"]
            cherry_pick_oids.add(merge_oid)
            # Include ancestor commits of merge commits to avoid false-positive warnings
            # for PRs that used a regular merge (not squash) strategy
            ancestor_output = run_command(["git", "log", "--format=%H", merge_oid, "--not", branch], silent=True)
            if ancestor_output:
                for ancestor_oid in ancestor_output.strip().splitlines():
                    cherry_pick_oids.add(ancestor_oid.strip())

    conflicting_prs_count = 0
    for pr in prs:
        if not pr.get("mergeCommit"):
            continue

        oid = pr["mergeCommit"]["oid"]
        number = pr["number"]

        files = get_changed_files(oid)
        if not files:
            continue

        # For each file, find commits that modified it between the target branch and the cherry-picked commit.
        # Deduplicate warnings: group affected files by missing commit.
        # missing_commits maps: missing_commit_oid -> {"title": ..., "files": [...]}
        missing_commits = {}

        for filepath in files:
            # git log <cherry-pick-commit> --not <target-branch> -- <file>
            output = run_command(["git", "log", oid, "--not", branch, "--format=%H %s", "--", filepath], silent=True)

            if not output:
                continue

            for line in output.strip().splitlines():
                parts = line.split(" ", 1)
                c = parts[0]
                title = parts[1] if len(parts) > 1 else ""

                if c == oid:
                    continue
                if c not in cherry_pick_oids:
                    entry = missing_commits.setdefault(c, {"title": title, "files": []})
                    if not entry["title"]:
                        entry["title"] = title
                    entry["files"].append(filepath)

        # Print deduplicated warnings
        if missing_commits:
            conflicting_prs_count += 1
            for missing_oid, entry in missing_commits.items():
                files_str = ", ".join(entry["files"])
                print(
                    f"WARNING: PR #{number} ({oid}) modifies files that were also changed by commit {missing_oid} ({entry['title']}), "
                    f"which is not in the cherry-pick list. This may indicate missing related changes. Affected files: {files_str}"
                )

    if conflicting_prs_count == 0:
        print("No potential missing dependencies found.")
    else:
        print(f"\nDone. Found potential missing dependencies for {conflicting_prs_count} PRs.")


def main():
    parser = argparse.ArgumentParser(description="Generate cherry-pick script from PRs with a specific label.")
    parser.add_argument("--label", required=True, help="Label to filter PRs")
    parser.add_argument(
        "--output", required=True, help="Output script file path (.sh for bash, .cmd for Windows batch)"
    )
    parser.add_argument("--repo", default="microsoft/onnxruntime", help="Repository (default: microsoft/onnxruntime)")
    parser.add_argument(
        "--branch", default="HEAD", help="Target branch to compare against for dependency checks (default: HEAD)"
    )
    parser.add_argument("--limit", type=int, default=200, help="Maximum number of PRs to fetch (default: 200)")
    parser.add_argument(
        "--md-output",
        help="Output markdown file path for the PR description (default: next to --output)",
    )
    args = parser.parse_args()

    # Preflight Check
    if not check_preflight():
        return

    # 1. Fetch Merged PRs
    prs = get_merged_prs(args.repo, args.label, args.limit)

    if not prs:
        print(f"No PRs found with label '{args.label}'.")
        return

    # Sort by mergedAt (ISO 8601 strings sort correctly in chronological order)
    prs.sort(key=lambda x: x["mergedAt"])

    # 1.5. Check which PRs are already in the target branch
    existing_prs = get_existing_pr_numbers(args.branch, repo=args.repo)
    if existing_prs:
        print(f"Found {len(existing_prs)} PRs already in branch '{args.branch}'.")

    cherry_pick_prs = []
    skipped_count = 0
    for pr in prs:
        number = pr["number"]
        safe_title = sanitize_title(pr["title"])

        if not pr.get("mergeCommit"):
            print(f"Warning: PR #{number} has no merge commit OID. Skipping.", file=sys.stderr)
            continue

        if number in existing_prs:
            print(f"Skipping PR #{number} (already in branch '{args.branch}'): {safe_title}")
            skipped_count += 1
            continue

        cherry_pick_prs.append(pr)

    # Determine output format based on file extension
    is_shell = args.output.endswith(".sh")

    # 2. Write Output Script
    commit_count = len(cherry_pick_prs)
    with open(args.output, "w", encoding="utf-8") as f:
        if is_shell:
            f.write("#!/bin/bash\n")
            f.write(f"# Cherry-pick {args.label} commits\n")
            f.write("# Sorted by merge time (oldest first)\n")
            f.write("set -e\n\n")
        else:
            f.write("@echo off\n")
            f.write(f"rem Cherry-pick {args.label} commits\n")
            f.write("rem Sorted by merge time (oldest first)\n\n")

        for pr in cherry_pick_prs:
            number = pr["number"]
            safe_title = sanitize_title(pr["title"])

            oid = pr["mergeCommit"]["oid"]
            comment = "#" if is_shell else "rem"
            f.write(f"{comment} PR {number}: {safe_title}\n")
            f.write(f"git cherry-pick {oid}\n\n")

    print(f"Generated {args.output} with {commit_count} commits ({skipped_count} skipped, already in branch).")

    # 3. Write PR Description Markdown (table format)
    output_dir = os.path.dirname(args.output)
    md_output = args.md_output or os.path.join(output_dir, "cherry_pick_pr_description.md")
    with open(md_output, "w", encoding="utf-8") as f:
        f.write("This cherry-picks the following commits for the release:\n\n")
        f.write("| Commit ID | PR Number | Commit Title |\n")
        f.write("|-----------|-----------|-------------|\n")
        for pr in cherry_pick_prs:
            number = pr["number"]
            title = escape_markdown_table_cell(pr["title"])
            oid = pr["mergeCommit"]["oid"]
            short_oid = oid[:10]
            f.write(f"| {short_oid} | #{number} | {title} |\n")

    print(f"Generated {md_output} with {commit_count} commits.")

    # 4. Dependency Check
    check_missing_dependencies(cherry_pick_prs, args.branch)


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
"""
Parse and Post customized perf data to DB
"""

import argparse
import csv
import datetime
import logging
import sys

import pandas as pd
from azure.kusto.data import KustoConnectionStringBuilder
from azure.kusto.data.data_format import DataFormat
from azure.kusto.ingest import IngestionProperties, QueuedIngestClient, ReportLevel

# Configure logging
LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)


def parse_mobile_perf(log_data: str, model: str, device_id: str, ep: str, commit_id: str, csv_filename: str):
    """
    Parse log data and save metrics to a CSV file.

    Args:
        log_data (str): The log data to parse.
        csv_filename (str): The filename to save the parsed metrics.
    """
    metrics = {
        "Model": model,
        "DeviceId": device_id,
        "Ep": ep,
        "CommitId": commit_id,
        "TTFTAvgTimeSec": None,
        "TTFTAvgTokenPerSec": None,
        "TokenGenerationAvgTimeSec": None,
        "TokenGenerationAvgTokenPerSec": None,
        "TokenSamplingAvgTimeSec": None,
        "TokenSamplingAvgTokenPerSec": None,
        "E2EGenerationAvgTimeSec": None,
        "PeakMemoryMB": None,
    }

    current_section = None

    for line in log_data.split("\n").strip():
        if "Prompt processing" in line:
            current_section = "TTFT"
        elif "Token generation" in line:
            current_section = "TokenGeneration"
        elif "Token sampling" in line:
            current_section = "TokenSampling"
        elif "E2E generation" in line:
            current_section = "E2EGeneration"
        elif "Peak working set size" in line:
            current_section = "PeakMemory"

        if line.startswith("avg (us):"):
            value = float(line.split(":")[1].strip()) / 1_000_000
            if current_section == "TTFT":
                metrics["TTFTAvgTimeSec"] = value
            elif current_section == "TokenGeneration":
                metrics["TokenGenerationAvgTimeSec"] = value
            elif current_section == "TokenSampling":
                metrics["TokenSamplingAvgTimeSec"] = value

        elif line.startswith("avg (tokens/s):"):
            value = float(line.split(":")[1].strip())
            if current_section == "TTFT":
                metrics["TTFTAvgTokenPerSec"] = value
            elif current_section == "TokenGeneration":
                metrics["TokenGenerationAvgTokenPerSec"] = value
            elif current_section == "TokenSampling":
                metrics["TokenSamplingAvgTokenPerSec"] = value

        elif line.startswith("avg (ms):"):
            value = float(line.split(":")[1].strip()) / 1_000
            if current_section == "E2EGeneration":
                metrics["E2EGenerationAvgTimeSec"] = value

        elif line.startswith("Peak working set size (bytes):"):
            value = float(line.split(":")[1].strip()) / (1024 * 1024)
            metrics["PeakMemoryMB"] = value

    for key in metrics.items():
        if metrics[key] is not None and isinstance(metrics[key], (int, float)):
            metrics[key] = f"{metrics[key]:.8f}"

    save_to_csv(metrics, csv_filename)


def save_to_csv(metrics: dict, csv_filename: str) -> None:
    try:
        with open(csv_filename, mode="w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(metrics.keys())
            writer.writerow(metrics.values())
        logging.info(f"Metrics saved to {csv_filename}")
    except OSError as e:
        logging.error(f"Failed to save metrics to {csv_filename}: {e}, abort.")
        sys.exit(1)


def post_to_db(csv_file: str, kusto_table: str, kusto_conn: str, kusto_db: str):
    """
    Post data to Kusto DB.

    Args:
        csv_file (str): The path to csv file.
        kusto_table (str): The Kusto table name.
        kusto_conn (str): The Kusto connection string.
        kusto_db (str): The Kusto database name.
    """
    try:
        table = pd.read_csv(csv_file)
        upload_time = datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0).isoformat()
        table["UploadTime"] = upload_time

        kcsb_ingest = KustoConnectionStringBuilder.with_az_cli_authentication(kusto_conn)
        ingest_client = QueuedIngestClient(kcsb_ingest)

        ingestion_props = IngestionProperties(
            database=kusto_db,
            table=kusto_table,
            data_format=DataFormat.CSV,
            report_level=ReportLevel.FailuresAndSuccesses,
        )

        ingest_client.ingest_from_dataframe(table, ingestion_properties=ingestion_props)
        logging.info(f"Data uploaded to Kusto table {kusto_table} in database {kusto_db}")
    except Exception as e:
        logging.error(f"Failed to upload data to Kusto DB: {e}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Parse and post perf data to DB")
    parser.add_argument("--kusto-table", required=True, help="Kusto table name")
    parser.add_argument("--kusto-conn", required=True, help="Kusto connection string")
    parser.add_argument("--kusto-db", required=True, help="Kusto database name")
    # Post mobile perf data
    parser.add_argument("--parse-mobile-perf", action="store_true", help="Parse mobile perf data and post to DB")
    parser.add_argument("--log-file", help="Path to log file containing performance data")
    parser.add_argument("--model", help="Testing model")
    parser.add_argument("--device-id", help="The local mobile device id")
    parser.add_argument("--commit-id", help="The ORT commit id")
    parser.add_argument("--ep", help="The execution provider running on devices")
    parser.add_argument("--output-csv", default="data.csv", help="CSV file to save parsed metrics")
    # Post csv data
    parser.add_argument("--upload-csv", help="CSV file to upload to DB")

    args = parser.parse_args()

    if args.parse_mobile_perf:
        for arg in [args.log_file, args.model, args.device_id, args.ep, args.commit_id]:
            if arg is None:
                raise ValueError(f"Missing required parameter {arg} for parsing mobile perf data")
        log_data = ""
        try:
            with open(args.log_file) as f:
                log_data = f.read()
            logging.info(f"Read mobile perf log from {args.log_file}")
        except OSError as e:
            logging.error(f"Failed to read log file {args.log_file}: {e}")
            sys.exit(1)
        # Parse the mobile perf data for further upload
        parse_mobile_perf(log_data, args.model, args.device_id, args.ep, args.commit_id, args.output_csv)
        post_to_db(args.output_csv, args.kusto_table, args.kusto_conn, args.kusto_db)
    elif args.upload_csv:
        # Upload an existing CSV to the database
        # Need to make sure schema is correct
        try:
            post_to_db(args.upload_csv, args.kusto_table, args.kusto_conn, args.kusto_db)
        except Exception as e:
            logging.error(f"Failed to upload CSV {args.upload_csv} to DB: {e}")
            sys.exit(1)
    else:
        parser.print_help()
        logging.error("Either --parse-mobile-perf or --upload-csv must be specified")
        sys.exit(1)
